/*
 * linux/arch/arm/mach-davinci/arch.c
 *
 * TI DaVinci Config file
 *
 * Copyright (C) 2004 Texas Instruments.
 *
 * ----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------------
 Modifications:
 ver. 1.0: Oct 2005, Swaminathan S
 -
 *
 */

/**************************************************************************
 * Included Files
 **************************************************************************/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/root_dev.h>

#include <asm/setup.h>
/*#include <asm/arch/ck.h>*/
#include <asm/io.h>
#include <asm/mach-types.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>

#include <asm/arch/hardware.h>


/**************************************************************************
 * Definitions
 **************************************************************************/
#define PINMUX0 __REG(0x01c40000)
#define PINMUX1 __REG(0x01c40004)
#define PTCMD   __REG(0x01C41120)
#define PDSTAT  __REG(0x01C41200)
#define PDCTL1  __REG(0x01C41304)
#define EPCPR   __REG(0x01C41070)
#define PTSTAT  __REG(0x01C41128)

#define MDSTAT  IO_ADDRESS(0x01C41800)
#define MDCTL   IO_ADDRESS(0x01C41A00)

/**************************************************************************
 * Public Functions
 **************************************************************************/
/**************************************
 Routine: board_setup_psc_on
 Description:  Enable a PSC domain
**************************************/

void board_setup_psc_on(unsigned int domain, unsigned int id)
{
	volatile unsigned int *mdstat = (unsigned int *)((int)MDSTAT + 4 * id);
	volatile unsigned int *mdctl = (unsigned int *)((int)MDCTL + 4 * id);

	*mdctl |= 0x00000003;	// Set PowerDomain to turn on

	if ((PDSTAT & 0x00000001) == 0) {
		PDCTL1 |= 0x1;
		PTCMD = (1 << domain);
		while ((((EPCPR >> domain) & 1) == 0)) ;

		PDCTL1 |= 0x100;
		while (!(((PTSTAT >> domain) & 1) == 0)) ;
	} else {
		PTCMD = (1 << domain);
		while (!(((PTSTAT >> domain) & 1) == 0)) ;
	}

	while (!((*mdstat & 0x0000001F) == 0x3)) ;
}

int cpu_type(void)
{
	return MACH_TYPE_DAVINCI_EVM;
}

extern void davinci_serial_init(void);
void board_init(void)
{
#ifdef CONFIG_BLK_DEV_DAVINCI
	/* Turn on the ATA PSC */
	board_setup_psc_on (DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_ATA);

	/* Enable ATA pin Mux */
	PINMUX0 |= (1 << 17) | (1 << 16);
#endif
#ifdef CONFIG_MMC_DAVINCI
	/* Turn on the ATA PSC */
	board_setup_psc_on (DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_MMC_SD);

	/* VDD power manupulations are done in U-Boot for CPMAC
	 * so applies to MMC as well
	 */
	PINMUX1 |= (~(1 << 9));
#endif
#if  (defined(CONFIG_I2C_DAVINCI) || defined(CONFIG_SOUND_DAVINCI))
	/* Turn on the ATA PSC */
	board_setup_psc_on (DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_I2C);

	PINMUX1 |= (1 << 7);
#endif
#ifdef CONFIG_SOUND_DAVINCI
	/* Turn on the I2C PSC */
	board_setup_psc_on(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_McBSP);

	/* Enable McBSP */
	PINMUX1 |= (1 << 10);
#endif
#ifdef CONFIG_VIDEO_DAVINCI
	board_setup_psc_on(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_VPSSMSTR);
	board_setup_psc_on(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_VPSSSLV);
	board_setup_psc_on(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_TPCC);
	board_setup_psc_on(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_TPTC0);
	board_setup_psc_on(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_TPTC1);
#endif
	davinci_serial_init();
#ifdef CONFIG_PM
	init_ck();
#endif
}

/*
 * DaVinci IO Mapping
 */
static struct map_desc davinci_io_desc[] __initdata = {
 { IO_VIRT,      	IO_PHYS,             IO_SIZE,        	   MT_DEVICE },
 { DAVINCI_IRAM_VIRT,   DAVINCI_IRAM_BASE,   SZ_16K,               MT_DEVICE },
};

void __init
davinci_map_io(void)
{
	iotable_init(davinci_io_desc, ARRAY_SIZE(davinci_io_desc));

	/* Initialize the DaVinci EVM board settigs */
	board_init ();
}

extern void davinci_irq_init(void);
extern struct sys_timer davinci_timer;

#if 0
MACHINE_START(DAVINCI_EVM, "DaVinci EVM")
    MAINTAINER("TI, PSP Team")
    BOOT_MEM(DAVINCI_DDR_BASE, IO_PHYS, IO_VIRT)
    BOOT_PARAMS(0x80000100)
    MAPIO(davinci_map_io)
    INITIRQ(davinci_irq_init)
    .timer = &davinci_timer,
    MACHINE_END
#else
MACHINE_START(DAVINCI_EVM, "DaVinci EVM")
//    MAINTAINER("TI, PSP Team")
//    BOOT_MEM(DAVINCI_DDR_BASE, IO_PHYS, IO_VIRT)

    .phys_ram	 = DAVINCI_DDR_BASE,
    .phys_io	 = IO_PHYS,
    .io_pg_offst = ((IO_VIRT)>>18)&0xfffc,

    .boot_params = 0x80000100,
    .map_io      = davinci_map_io,
    .init_irq    = davinci_irq_init,
    .timer       = &davinci_timer,
MACHINE_END
#endif
EXPORT_SYMBOL(cpu_type);




