/*
 * linux/arch/arm/mach-davinci/dma.c
 *
 * TI DaVinci DMA file
 *
 * Copyright (C) 2004 Texas Instruments.
 *
 * ----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------------
 Modifications:
 ver. 1.0: Oct 2005, Swaminathan S
 -
 *
 */

#include <asm/arch/edma.h>
#include "dma_osdep.h"

//#define EDMA_DEBUG

/* Structure containing the dma channel parameters */
static struct davinci_dma_lch {
    int dev_id;
    int inUse; /*1-used 0-unused*/
    //void *data;
    //void (*callback) (int lch, unsigned short ch_status, void *data);
    int link_lch;
    int dma_running;
    int param_no;
    int tcc;
}dma_chan[DAVINCI_EDMA_NUM_PARAMENTRY];

static struct dma_interrupt_data{
    void (*callback) (int lch, unsigned short ch_status, void *data);
    void *data;
}intr_data[64];

/*
 *  Each bit field of the elements bellow indicate the corresponding EDMA channel
 *  availability  on arm side events
 */
static unsigned long edma_channels_arm[] = {
    0xffffffff,
    0xffffffff
    };

/*
    Each bit field of the elements bellow indicate the corresponding QDMA channel
    availability  on arm side events
*/
static unsigned char qdma_channels_arm[] = {
    0xff
    };

/*
   Each bit field of the elements bellow indicate corresponding PARAM entry
   availibility on arm side events
*/
static unsigned long param_entry_arm[] = {
    0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
    0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
    0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
    0xffffffff};

/*
   Each bit field of the elements bellow indicate whether a PARAM entry
   is free or in use
   1 - free
   0 - in use
*/
static unsigned long param_entry_use_status[] = {
    0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
    0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
    0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
    0xffffffff};

/*
   Each bit field of the elements bellow indicate whether a intrerrupt
   is free or in use
   1 - free
   0 - in use
*/
static unsigned long dma_intr_use_status[] = {
    0xffffffff,
    0xffffffff
    };

/*
    This lists the DMA channel numbers which does not have any events
    associated with it
*/
static int dmachan_noevent[] = {
    0, 1, 12, 13, 14, 15, 25, 30, 31, 45, 46, 47, 55, 56, 57, 58, 59, 60,
    61, 62, 63, -1};

static int channel_queue_mapping[][2] =  {
/* {channel no, event queue no } */
       {0,0},  {1,1},  {2,0},  {3,1},  {4,0},  {5,1},  {6,0},  {7,1},
       {8,0},  {9,1}, {10,0}, {11,1}, {12,0}, {13,1}, {14,0}, {15,1},
      {16,0}, {17,1}, {18,0}, {19,1}, {20,0}, {21,1}, {22,0}, {23,1},
      {24,0}, {25,1}, {26,0}, {27,1}, {28,0}, {29,1}, {30,0}, {31,1},
      {32,0}, {33,1}, {34,0}, {35,1}, {36,0}, {37,1}, {38,0}, {39,1},
      {40,0}, {41,1}, {42,0}, {43,1}, {44,0}, {45,1}, {46,0}, {47,1},
      {48,0}, {49,1}, {50,0}, {51,1}, {52,0}, {53,1}, {54,0}, {55,1},
      {56,0}, {57,1}, {58,0}, {59,1}, {60,0}, {61,1}, {62,0}, {63,1},
      {64,0}, {65,1}, {66,0}, {67,1}, {68,0}, {69,1}, {70,0}, {71,1},
      {-1, -1}
/*end of mapping*/
};

static int queue_tc_mapping[DAVINCI_EDMA_NUM_EVQUE + 1][2] =  {
/* {event queue no, TC no} */
    {0,0},
    {1,1},
    {-1,-1}
};

static int queue_priority_mapping[DAVINCI_EDMA_NUM_EVQUE + 1][2] =  {
    /* {event queue no, Priority} */
    {0,0},
    {1,1},
    {-1,-1}
    };

static int qdam_to_param_mapping[8] =  {0};

static volatile edmacc_regs *ptr_edmacc_regs = NULL;


/*****************************************************************************/


static void memcpylocal(void *dst, void* src, unsigned int len)
{
	unsigned int i=0;
	unsigned char *sr = (unsigned char *)src;
	unsigned char *ds = (unsigned char *)dst;
	if((sr == NULL) || (ds==NULL))
	return;
	for( i=0;i<len;i++)
	{
		*ds=*sr;
		ds++;
		sr++;
	}
}


static void memsetlocal(void *dst, unsigned char data, unsigned int len)
{
	unsigned int i=0;
	unsigned char *ds = (unsigned char *)dst;
	if(ds==NULL)
	return;
	for( i=0;i<len;i++)
	{
		*ds=data;
		ds++;
	}
}

static void map_dmach_queue(int ch_no, int queue_no)
{
    if(ch_no < DAVINCI_EDMA_NUM_DMACH)
    {
	int bitStart = (ch_no % 8) * 4;
	ptr_edmacc_regs->dmaqnum[ch_no >> 3] &= (~(0x7 << bitStart));
	ptr_edmacc_regs->dmaqnum[ch_no >> 3] |= ((queue_no & 0x7) << bitStart);
    }
    else if(ch_no >= DAVINCI_EDMA_NUM_DMACH && ch_no < (DAVINCI_EDMA_NUM_DMACH + \
	DAVINCI_EDMA_NUM_QDMACH))
    {
	int bitStart = (ch_no - DAVINCI_EDMA_NUM_DMACH) * 4;
	ptr_edmacc_regs->qdmaqnum &= (~(0x7 << bitStart));
	ptr_edmacc_regs->qdmaqnum |= ((queue_no & 0x7) << bitStart);
    }
}


/* For Davinci this Macro supports mapping only for QDMA channels and PaRam
   entry */
static void map_dmach_param(int ch_no,int param_no)
{
    if(ch_no >= DAVINCI_EDMA_NUM_DMACH && ch_no < (DAVINCI_EDMA_NUM_DMACH + \
	DAVINCI_EDMA_NUM_QDMACH))
    {
	ptr_edmacc_regs->qchmap[ch_no-DAVINCI_EDMA_NUM_DMACH] &= \
	    ~(PAENTRY|TRWORD);
	ptr_edmacc_regs->qchmap[ch_no-DAVINCI_EDMA_NUM_DMACH] |= \
	    (((param_no&0x1ff) << 5)|(QDMA_TRWORD<<2));
    }
}


static void map_queue_tc(int queue_no,int tc_no)
{
    int bitstart = queue_no * 4;
    ptr_edmacc_regs->quetcmap &= ~(0x7 << bitstart);
    ptr_edmacc_regs->quetcmap |= ((tc_no & 0x7) << bitstart);
}

static void assign_priority_to_queue(int queue_no, int priority)
{
    int bitstart = queue_no * 4;
    ptr_edmacc_regs->quepri &= ~(0x7 << bitstart);
    ptr_edmacc_regs->quepri |= ((priority & 0x7) << bitstart);
}

/******************************************************************************
 *
 * DMA Param entry requests: Requests for the param structure entry for the dma
 *                          channel passed
 * Arguments:
 *      lch  - logical channel for which param entry is being requested.
 *
 * Return: param number on success, or negative error number on failure
 *
 *****************************************************************************/
static int request_param(int lch,int dev_id)
{
    int i=0, j=0, isbreak = 0;

    if(lch >=0 && lch <DAVINCI_EDMA_NUM_DMACH )
    {
	/*In davinci there is 1:1 mapping between edma channels and parma nums*/
	param_entry_use_status[lch/32] &= (~(1<<(lch%32)));
#ifdef EDMA_DEBUG
	/*DEBUGMSG("param no=%d\r\n",lch);*/
#endif
	return lch;
    }
    else
    {
	if(dev_id >= DAVINCI_DMA_QDMA0 && dev_id <= DAVINCI_DMA_QDMA7)
	{
	    i = 0;
	}
	else if(dev_id == DAVINCI_EDMA_PARAM_ANY)
	{
	    i = DAVINCI_EDMA_NUM_DMACH/*+DAVINCI_EDMA_NUM_QDMACH*/;
	}

	while(i<DAVINCI_EDMA_NUM_PARAMENTRY)
	{
	    j=0,isbreak = 1;
	    if((param_entry_arm[i/32] & (1<<(i%32))) &&
		(param_entry_use_status[i/32] & (1<<(i%32))))
	    {
		 if(dev_id != DAVINCI_EDMA_PARAM_ANY)
		 {
		     while(dmachan_noevent[j] != -1)
		     {
			 if(dmachan_noevent[j] == i)
			 {
			     isbreak = 0;
			 }
			 j++;
		     }

		     if(!isbreak)
		     {
			 break;
		     }
		 }
		 else
		 {
		     break;
		 }
		 i++;
	    }
	    else
	    {
		i++;
	    }
	}

	if(i<DAVINCI_EDMA_NUM_PARAMENTRY)
	{
	    param_entry_use_status[i/32] &= (~(1<<(i%32)));
	    DEBUGMSG("param no=%d\r\n",i);
	    return i;  /*return the param number which is free*/
	}
	else
	    return -1; /*no free param*/
    }
}

/******************************************************************************
 *
 * Free dma param entry: Freethe param entry number passed
 * Arguments:
 *      param_no - Param entry to be released or freed out
 *
 * Return: N/A
 *
 *****************************************************************************/
static void free_param(int param_no)
{
    if(param_no>=0 && param_no <DAVINCI_EDMA_NUM_PARAMENTRY)
	param_entry_use_status[param_no/32] |= (1<<(param_no%32));
}

/******************************************************************************
 *
 * DMA interrupt requests: Requests for the interrupt on the free channel
 *
 * Arguments:
 *      lch - logical channel number for which the interrupt is to be requested
 *            for the free channel.
 *      callback - callback function registered for the requested interrupt
 *                 channel
 *      data - channel private data.
 *
 * Return: free interrupt channel number on success, or negative error number
 *              on failure
 *
 *****************************************************************************/
static int request_dma_interrupt(int lch,
	      void (*callback) (int lch, unsigned short ch_status, void *data),
	      void *data,int paramno,int requested_tcc)
{
    signed int free_intr_no = -1;
    int i = 0, j = 0, isbreak = 0;

    /*edma channels*/
    if(lch >= 0 && lch < DAVINCI_EDMA_NUM_DMACH)
    {
	if(dma_intr_use_status[lch/32] & (1<<(lch%32)))
	{
	    dma_intr_use_status[lch/32] &= (~(1<<(lch%32))); /*in use*/
	    free_intr_no = lch;
#ifdef EDMA_DEBUG
	    DEBUGMSG("interrupt no=%d\r\n",free_intr_no);
#endif
	}
	else
	{
	    DEBUGMSG("EDMA:Error\r\n");
	    return -1;
	}
    }
    /*qdma channels*/
    else if(lch >= DAVINCI_EDMA_NUM_DMACH &&
	    lch < (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	if(requested_tcc != TCC_ANY)
	{
	    if(dma_intr_use_status[requested_tcc/32] & (1<<(requested_tcc%32)))
	    {
		j=0;isbreak=1;
		while(dmachan_noevent[j] != -1)
		{
		    if(dmachan_noevent[j] == requested_tcc)
		    {
			isbreak = 0;
			break;
		    }
		    j++;
		}

		if(!isbreak)
		{
		    dma_intr_use_status[requested_tcc/32] &= (~(1<<(requested_tcc%32)));
		    free_intr_no = requested_tcc;
#ifdef EDMA_DEBUG
		    DEBUGMSG("interrupt no=%d\r\n",free_intr_no);
#endif
		}
		else
		{
		    DEBUGMSG("EDMA:Error - wrong tcc passed\r\n");
		    return -1;
		}

	    }
	    else
	    {
		DEBUGMSG("EDMA:Error - wrong tcc passed\r\n");
		return -1;
	    }
	}
	else
	{
	    i= 0;
	    while(i < DAVINCI_EDMA_NUM_DMACH)
	    {
		j=0; isbreak=1;
		if(dma_intr_use_status[i/32] & (1<<(i%32)))
		{
		    while(dmachan_noevent[j] != -1)
		    {
			if(dmachan_noevent[j] == i)
			{
			    isbreak = 0;
			    break;
			}
			j++;
		    }

		    if(!isbreak)
		    {
			dma_intr_use_status[i/32] &= (~(1<<(i%32)));
			free_intr_no = i;
#ifdef EDMA_DEBUG
			DEBUGMSG("interrupt no=%d\r\n",free_intr_no);
#endif
			break;
		    }
		    i++;
		}
		else
		{
		     i++;
		}
	    }
	}
    }
    else
    {
	DEBUGMSG("EDMA : ERROR lch = %d\r\n",lch);
    }

    if(isbreak)
    {
	DEBUGMSG("EDMA : while allocating EDMA channel for QDMA");
    }

    if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
	(DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	if(free_intr_no < 32)
	{
	    ptr_edmacc_regs->dra[0].drae = ptr_edmacc_regs->dra[0].drae | (1<<free_intr_no);
	}
	else
	{
	    ptr_edmacc_regs->dra[0].draeh = ptr_edmacc_regs->dra[0].draeh | (1<< (free_intr_no-32));
	}
    }

    if(free_intr_no >= 0 && free_intr_no < 64)
    {
	(free_intr_no < 32) ? (ptr_edmacc_regs->shadow[0].iesr |= \
		(1UL << (free_intr_no))): (ptr_edmacc_regs->shadow[0].iesrh |= \
		    (1UL << (free_intr_no-32)));
	intr_data[free_intr_no].callback = callback;
	intr_data[free_intr_no].data = data;
    }
	return free_intr_no;
}

/******************************************************************************
 *
 * Free the dma interrupt: Releases the dma interrupt on the channel
 *
 * Arguments:
 *      intr_no - interrupt number on the channel to be released or freed out
 *
 * Return: N/A
 *
 *****************************************************************************/
static void free_dma_interrupt(int intr_no)
{
    if(intr_no >= 0 && intr_no < 64)
    {
	(intr_no<32)?(ptr_edmacc_regs->shadow[0].icr |= \
		(1UL << (intr_no))): (ptr_edmacc_regs->shadow[0].icrh |= \
		    (1UL << (intr_no-32)));
	dma_intr_use_status[intr_no/32] |= (1<<(intr_no%32)); /*free*/
	intr_data[intr_no].callback = NULL;
	intr_data[intr_no].data = NULL;
    }
}

/******************************************************************************
 *
 * DMA interrupt handler
 *
 *****************************************************************************/
static void dma_irq_handler(void)
{
    int i;
//    unsigned long  eIRQ0,eIRQ1;
    unsigned int cnt;

    cnt=0;
    if((ptr_edmacc_regs->shadow[0].ipr ==0 ) && (ptr_edmacc_regs->shadow[0].iprh ==0 ))
	return ;

    while(1)
    {
	if(ptr_edmacc_regs->shadow[0].ipr)
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("IPR =%d\r\n",ptr_edmacc_regs->shadow[0].ipr);
#endif
	    for(i=0; i<32;i++)
	    {
		if(ptr_edmacc_regs->shadow[0].ipr & (1 << i))
		{
		    /* here write to ICR to clear the corresponding IPR bits*/
		    ptr_edmacc_regs->shadow[0].icr |= (1<<i);
		    if(intr_data[i].callback)
		    {
			intr_data[i].callback(i, DMA_COMPLETE, intr_data[i].data);
		    }
		}
	    }
	}
	else if(ptr_edmacc_regs->shadow[0].iprh)
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("IPRH =%d\r\n",ptr_edmacc_regs->shadow[0].iprh);
#endif
	    for(i=0; i<32;i++)
	    {
		if(ptr_edmacc_regs->shadow[0].iprh & (1 << i))
		{
		    /* here write to ICR to clear the corresponding IPR bits*/
		    ptr_edmacc_regs->shadow[0].icrh |= (1<<i);
		    if(intr_data[i].callback)
		    {
			intr_data[32+i].callback(32+i, DMA_COMPLETE, intr_data[32+i].data);
		    }
		}
	    }
	}

	if((ptr_edmacc_regs->shadow[0].ipr ==0 ) && (ptr_edmacc_regs->shadow[0].iprh ==0 ))
	{
	     break ;
	}

	cnt++;
	if(cnt > 10)
	{
	     break;
	}

    }

    ptr_edmacc_regs->shadow[0].ieval=0x1;
}

/******************************************************************************
 *
 * DMA error interrupt handler
 *
 *****************************************************************************/
static void dma_ccerr_handler(void)
{

    int i;
//    unsigned long  eIRQ0,eIRQ1;
    unsigned int cnt;

    cnt=0;
    if((ptr_edmacc_regs->emr ==0 ) && (ptr_edmacc_regs->emr ==0 ) &&
       (ptr_edmacc_regs->qemr==0) && (ptr_edmacc_regs->ccerr==0))
	return ;

    while(1)
    {
	if(ptr_edmacc_regs->emr)
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("EMR =%d\r\n",ptr_edmacc_regs->emr);
#endif
	    for(i=0; i<32;i++)
	    {
		if(ptr_edmacc_regs->emr & (1 << i))
		{
		    /* here write to EMCR to clear the corresponding EMR bits*/
		    ptr_edmacc_regs->emcr |= (1<<i);
		    /*Clear any SER*/
		    ptr_edmacc_regs->shadow[0].secr |= (1<<i);
		    if(intr_data[i].callback)
		    {
			intr_data[i].callback(i, DMA_CC_ERROR, intr_data[i].data);
		    }

		    /*      ptr_edmacc_regs->CCERRCLR |=((1<<16)|0x3);*/
		}
	    }
	}
	else if(ptr_edmacc_regs->emrh)
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("EMRH =%d\r\n",ptr_edmacc_regs->emrh);
#endif
	    for(i=0; i<32;i++)
	    {
		if(ptr_edmacc_regs->emrh & (1 << i))
		{
		    /* here write to ICR to clear the corresponding IPR bits*/
		    ptr_edmacc_regs->emcrh |= (1<<i);
		    /*Clear any SER*/
		    ptr_edmacc_regs->shadow[0].secrh |= (1<<i);
		    if(intr_data[i].callback)
		    {
			intr_data[i].callback(i, DMA_CC_ERROR, intr_data[i].data);
		    }

		    /*      ptr_edmacc_regs->CCERRCLR |=((1<<16)|0x3);*/
		}
	    }
	}
	else if(ptr_edmacc_regs->qemr)
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("QEMR =%d\r\n",ptr_edmacc_regs->qemr);
#endif
	    for(i=0; i<8;i++)
	    {
		if(ptr_edmacc_regs->qemr & (1 << i))
		{
		    /* here write to ICR to clear the corresponding IPR bits*/
		    ptr_edmacc_regs->qemcr |= (1<<i);
		    ptr_edmacc_regs->shadow[0].qsecr |= (1<<i);
		}
	    }
	}
	else if(ptr_edmacc_regs->ccerr)
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("CCERR =%d\r\n",ptr_edmacc_regs->ccerr);
#endif
	    for(i=0; i<8;i++)
	    {
		if(ptr_edmacc_regs->ccerr & (1 << i))
		{
		    /* here write to ICR to clear the corresponding IPR bits*/
		    ptr_edmacc_regs->ccerrclr |= (1<<i);
		}
	    }
	}

	if((ptr_edmacc_regs->emr ==0 ) && (ptr_edmacc_regs->emrh ==0 ) &&
	    (ptr_edmacc_regs->qemr==0) && (ptr_edmacc_regs->ccerr==0))
	{
	    break ;
	}

	cnt++;
	if(cnt > 10)
	{
	    break;
	}
    }

    ptr_edmacc_regs->eeval = 0x1;
}

/******************************************************************************
 *
 * DMA error interrupt handler
 *
 *****************************************************************************/
static void dma_tc1err_handler(void)
{
}

/******************************************************************************
 *
 * DMA error interrupt handler
 *
 *****************************************************************************/
static void dma_tc2err_handler(void)
{
}

/******************************************************************************
 *
 * DMA initialisation on davinci
 *
 *****************************************************************************/
static int __init arch_dma_init(void)
{
    int i;

    ptr_edmacc_regs = get_edma_base();

    DEBUGMSG("DMA REG BASE ADDR=%x\n",(unsigned)ptr_edmacc_regs);


    memsetlocal(dma_chan, 0x00, sizeof(dma_chan));


    memsetlocal( (void *)&(ptr_edmacc_regs->paramentry[0]) ,0x00, \
	sizeof(ptr_edmacc_regs->paramentry));

    i = 0;

    /* Channel to queue mapping */
    while (channel_queue_mapping[i][0] != -1)
    {
	map_dmach_queue(channel_queue_mapping[i][0],
			channel_queue_mapping[i][1]);
	i++;
    }

    i = 0;

    /* Event queue to TC mapping */
    while (queue_tc_mapping[i][0] != -1)
    {
	map_queue_tc(queue_tc_mapping[i][0], queue_tc_mapping[i][1]);
	i++;
    }

    i = 0;

    /* Event queue priority mapping */
    while (queue_priority_mapping[i][0] != -1)
    {
	assign_priority_to_queue(queue_priority_mapping[i][0],
				 queue_priority_mapping[i][1]);
	i++;
    }

    for (i = 0; i < DAVINCI_EDMA_NUM_REGIONS; i++)
    {
	ptr_edmacc_regs->dra[i].drae = 0x0;
	ptr_edmacc_regs->dra[i].draeh = 0x0;
	ptr_edmacc_regs->qrae[i] = 0x0;
    }
    LOCK_INIT;

    return 0;
}


/******************************************************************************
 *
 * DMA channel requests: Requests for the dma device passed if it is free
 *
 * Arguments:
 *      dev_id     - request for the param entry device id
 *      dev_name   - device name
 *      callback   - pointer to the channel callback.
 *      Arguments:
 *          lch  - channel no, which is the IPR bit position, indicating from which
 *                 channel the interrupt arised.
 *          data - channel private data, which is received as one of the arguments
 *                 in davinci_request_dma.
 *      data - private data for the channel to be requested, which is used to
 *                   pass as a parameter in the callback function in irq handler.
 *      lch - contains the device id allocated
 *  tcc        - Transfer Completion Code, used to set the IPR register bit
 *                   after transfer completion on that channel.
 *  eventq_no  - Event Queue no to which the channel will be associated with
 *               (valied only if you are requesting for a DMA MasterChannel)
 *               Values : 0 to 7
 *                       -1 for Default queue
 * INPUT:   dev_id
 * OUTPUT:  *dma_ch_out
 *
 * Return: zero on success, or corresponding error no on failure
 *
 *****************************************************************************/
int davinci_request_dma(int dev_id, const char *dev_name,
    void (*callback) (int lch, unsigned short ch_status, void *data),
	void *data, int *lch, int *tcc, enum dma_event_q eventq_no)
{
    int retval = 0, i=0; //,param_no;
    static int reqflag=0;
    int temp_ch=0;

    /* checking the ARM side events */
    if (dev_id >= 0 && (dev_id < DAVINCI_EDMA_NUM_DMACH))
    {
	if (!(edma_channels_arm[dev_id/32] & (0x1 << (dev_id%32))))
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("dev_id = %d not supported on ARM side\r\n", dev_id);
#endif
	    return PSP_E_NOT_SUPPORTED;
	}
    }
    else if(dev_id >= DAVINCI_EDMA_NUM_DMACH && dev_id <=(DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {

       if ( !(qdma_channels_arm[0] & (0x1 << (dev_id-DAVINCI_EDMA_NUM_DMACH))))
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("dev_id = %d not supported on ARM side\r\n", dev_id);
#endif
	    return PSP_E_RESOURCES;
	}
    }


	if((dev_id != DAVINCI_DMA_CHANNEL_ANY) && (dev_id !=DAVINCI_EDMA_PARAM_ANY))
	{
		if (dev_id >= DAVINCI_EDMA_NUM_DMACH && dev_id <(DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
		{
			ptr_edmacc_regs->qrae[0] = ptr_edmacc_regs->qrae[0] | ( 1 << (dev_id - DAVINCI_EDMA_NUM_DMACH));
		}
		else
		{
			if(dev_id < 32)
			{
			    ptr_edmacc_regs->dra[0].drae = ptr_edmacc_regs->dra[0].drae | (1<<dev_id);
			}
			else
			{
			ptr_edmacc_regs->dra[0].draeh = ptr_edmacc_regs->dra[0].draeh | (1<< (dev_id-32));
			}
		}
	}

    LOCK;
    //PAL_osProtectEntry(PAL_OSPROTECT_INTERRUPT, &intstate);  /* ## */
/*move to init*/
    if (!reqflag)
    {
	if(register_dma_interrupts(dma_irq_handler, dma_ccerr_handler,\
	    dma_tc1err_handler, dma_tc2err_handler))
	{
#ifdef EDMA_DEBUG
	    DEBUGMSG("register_dma_interrupts failed\r\n");
#endif
	    UNLOCK;
	   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
	    return PSP_E_RESOURCES;
	}
	else
	    reqflag = 1;
    }
/*move to init*/

    if (dev_id >= 0 && dev_id < (DAVINCI_EDMA_NUM_DMACH))
    {

	/* Since the first 64 Channels are mapped to the first 64 PARAM entries */
	if (!dma_chan[dev_id].inUse)
	{
	    *lch = dev_id;
	    dma_chan[*lch].param_no = request_param(*lch,dev_id);
	    if(dma_chan[*lch].param_no == -1)
	    {
	       //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
	       UNLOCK;
		return PSP_E_RESOURCES;
	    }
	    else
#ifdef EDMA_DEBUG
	       DEBUGMSG( "param_no=%d\r\n",dma_chan[*lch].param_no);
#endif
	    if(callback)
	    {
		dma_chan[*lch].tcc = request_dma_interrupt(*lch,callback,data,dma_chan[*lch].param_no,*tcc);
		if(dma_chan[*lch].tcc == -1)
		{
		   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
		   UNLOCK;
		    return PSP_E_RESOURCES;
		}
		else
		{
		    *tcc = dma_chan[*lch].tcc;
#ifdef EDMA_DEBUG
		    DEBUGMSG( "tcc_no=%d\r\n",dma_chan[*lch].tcc);
#endif
		}
	    }
	    else
		dma_chan[*lch].tcc = -1;

	    map_dmach_queue(dev_id, eventq_no);

	    retval = 0;
	}
	else
	    retval = PSP_E_RESOURCES;
    }
    else if (dev_id >= DAVINCI_EDMA_NUM_DMACH && dev_id < \
	(DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	if (!dma_chan[qdam_to_param_mapping[dev_id-DAVINCI_EDMA_NUM_DMACH]].inUse)
	{
	    *lch = dev_id;

	    dma_chan[*lch].param_no = request_param(*lch,dev_id);
	    if(dma_chan[*lch].param_no == -1)
	    {
#ifdef EDMA_DEBUG
		DEBUGMSG("request_param failed\r\n");
#endif
	       //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
	       UNLOCK;
	       return PSP_E_RESOURCES;
	    }
	    else
	    {
#ifdef EDMA_DEBUG
	      DEBUGMSG( "param_no=%d\r\n",dma_chan[*lch].param_no);
#endif
		map_dmach_param(*lch,dma_chan[*lch].param_no);
	    }

	    if(callback)
	    {
		dma_chan[*lch].tcc = request_dma_interrupt(*lch,callback,data,dma_chan[*lch].param_no,*tcc);
		if(dma_chan[*lch].tcc == -1)
		{
		   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
		   UNLOCK;
		    return PSP_E_RESOURCES;
		}
		else
		{
		    *tcc = dma_chan[*lch].tcc;
#ifdef EDMA_DEBUG
		    DEBUGMSG( "tcc_no=%d\r\n",dma_chan[*lch].tcc);
#endif
		}
	    }
	    else
		    dma_chan[*lch].tcc = -1;

	    map_dmach_queue(dev_id, eventq_no);
	    retval = 0;
	}
	else
	    retval = PSP_E_RESOURCES;

    }
    else if (dev_id == DAVINCI_DMA_CHANNEL_ANY)
    {
	i = 0;
	retval = 0;
	while (dmachan_noevent[i] != -1)
	{
	    if (!dma_chan[dmachan_noevent[i]].inUse)
	    {
		*lch = dmachan_noevent[i];
#ifdef EDMA_DEBUG
		DEBUGMSG("any chan = %x\r\n",*lch);
#endif
		dma_chan[*lch].param_no = request_param(*lch,dev_id);
		if(dma_chan[*lch].param_no == -1)
		{
		   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
		   UNLOCK;
		    return PSP_E_RESOURCES;
		}

#ifdef EDMA_DEBUG
		    DEBUGMSG( "param_no=%d\r\n",dma_chan[*lch].param_no);
#endif
		if (dma_chan[*lch].param_no >= DAVINCI_EDMA_NUM_DMACH && dma_chan[*lch].param_no <(DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
		{
		    ptr_edmacc_regs->qrae[0] =        ptr_edmacc_regs->qrae[0] | ( 1 << (dma_chan[*lch].param_no - DAVINCI_EDMA_NUM_DMACH));
		}
		else
		{
		    if(dma_chan[*lch].param_no < 32)
		    {
			ptr_edmacc_regs->dra[0].drae = ptr_edmacc_regs->dra[0].drae | (1<<dma_chan[*lch].param_no);
		    }
		    else
		    {
			ptr_edmacc_regs->dra[0].draeh = ptr_edmacc_regs->dra[0].draeh | (1<< (dma_chan[*lch].param_no-32));
		    }
		}

		if(callback)
		{
		    dma_chan[*lch].tcc = request_dma_interrupt(*lch,callback,data,dma_chan[*lch].param_no,*tcc);

		    if(dma_chan[*lch].tcc == -1)
		    {
		       //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
		       UNLOCK;
			return PSP_E_RESOURCES;
		    }
		    else
		    {
			*tcc = dma_chan[*lch].tcc;
#ifdef EDMA_DEBUG
			/*DEBUGMSG( "tcc_no=%x\r\n",dma_chan[*lch].tcc);*/
#endif
		    }
		}
		else
		{
		    dma_chan[*lch].tcc = -1;
		}

		map_dmach_queue(dev_id, eventq_no);
		retval = 0;
		break;
	    }
	    i++;
	}
    }
    else if (dev_id == DAVINCI_EDMA_PARAM_ANY)
    {
	retval = 0;
	for (i = (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH);
		      i < DAVINCI_EDMA_NUM_PARAMENTRY; i++)
	{
	    if (!dma_chan[i].inUse)
	    {
#ifdef EDMA_DEBUG
		DEBUGMSG("any link = %d\r\n",i);
#endif
		*lch = i;

		dma_chan[*lch].param_no = request_param(*lch,dev_id);
		if(dma_chan[*lch].param_no == -1)
		{
		    DEBUGMSG("request_param failed\r\n");
		   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
		   UNLOCK;
		    return PSP_E_RESOURCES;
		}
		else
		{
#ifdef EDMA_DEBUG
		    DEBUGMSG( "param_no=%d\r\n",dma_chan[*lch].param_no);
#endif
		}
		if(*tcc != -1)
		    dma_chan[*lch].tcc = *tcc;
		else
		    dma_chan[*lch].tcc = -1;

		retval = 0;
		break;
	    }
	}
    }
    else
    {
	retval = PSP_E_INVAL_STATE;
    }

    if (!retval)
    {
	if (dev_id >= DAVINCI_EDMA_NUM_DMACH && dev_id < \
	    (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{
		/*Master Channel*/
	    qdam_to_param_mapping[dev_id-DAVINCI_EDMA_NUM_DMACH] = dma_chan[*lch]. param_no;
	    dma_chan[qdam_to_param_mapping[dev_id-DAVINCI_EDMA_NUM_DMACH]].inUse = 1;
	    dma_chan[qdam_to_param_mapping[dev_id-DAVINCI_EDMA_NUM_DMACH]].dev_id = *lch;
	    dma_chan[qdam_to_param_mapping[dev_id-DAVINCI_EDMA_NUM_DMACH]].tcc = dma_chan[*lch].tcc;

		temp_ch = qdam_to_param_mapping[dev_id-DAVINCI_EDMA_NUM_DMACH];
	    if (dma_chan[*lch].tcc != -1)
	    {
		ptr_edmacc_regs->paramentry[dma_chan[temp_ch].param_no].opt &= (~TCC);
		ptr_edmacc_regs->paramentry[dma_chan[temp_ch].param_no].opt |= \
			((0x3f & dma_chan[*lch].tcc) << 12);
		/*set TCINTEN bit in PARAM entry*/
		ptr_edmacc_regs->paramentry[dma_chan[temp_ch].param_no].opt |= TCINTEN;
	    }
	    else
	    {
		ptr_edmacc_regs->paramentry[dma_chan[temp_ch].param_no].opt &= ~TCINTEN;
	    }

	    /* assign the link field to no link. i.e 0xffff  */
	    ptr_edmacc_regs->paramentry[dma_chan[temp_ch].param_no].link_bcntrld |= 0xffff;
	}
	else
	{
	    /*Slave Channel*/
	    dma_chan[*lch].inUse = 1;
	    dma_chan[*lch].dev_id = *lch;
	    if (dma_chan[*lch].tcc != -1)
	    {
		ptr_edmacc_regs->paramentry[dma_chan[*lch].param_no].opt &= (~TCC);
		ptr_edmacc_regs->paramentry[dma_chan[*lch].param_no].opt |= \
		((0x3f & dma_chan[*lch].tcc) << 12);
		/*set TCINTEN bit in PARAM entry*/
		ptr_edmacc_regs->paramentry[dma_chan[*lch].param_no].opt |= TCINTEN;
	    }
	    else
	    {
		ptr_edmacc_regs->paramentry[dma_chan[*lch].param_no].opt &= ~TCINTEN;
	    }

	    /* assign the link field to no link. i.e 0xffff  */
	    ptr_edmacc_regs->paramentry[dma_chan[*lch].param_no].link_bcntrld |= 0xffff;
	}
    }
    UNLOCK;
   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
    return retval;
}

/******************************************************************************
 *
 * DMA channel free: Free dma channle
 * Arguments:
 *      dev_id     - request for the param entry device id
 *
 * Return: zero on success, or corresponding error no on failure
 *
 *****************************************************************************/
void davinci_free_dma (int lch)
{
	int temp_ch = 0;

	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    LOCK;
    //PAL_osProtectEntry(PAL_OSPROTECT_INTERRUPT, &intstate);  /* ## */
    dma_chan[lch].inUse = 0;
    free_param(dma_chan[lch].param_no);
	if(lch>=0 && lch<(DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	free_dma_interrupt(dma_chan[lch].tcc);

    UNLOCK;
   //PAL_osProtectExit(PAL_OSPROTECT_INTERRUPT,intstate); /* ## */
}


/******************************************************************************
 *
 * DMA source parameters setup
 * ARGUMENTS:
 *      lch         - channel for which the source parameters to be configured
 *      src_port    - Source port address
 *      addressMode - indicates wether addressing mode is fifo.
 *
 *****************************************************************************/
void davinci_set_dma_src_params(int lch, unsigned long src_port, \
    enum address_mode mode, enum fifo_width width)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	/* set the source port address in source register of param structure */
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src = src_port;


	/* set the fifo addressing mode */
	if (mode)
	{   /*reset SAM and FWID*/
	    ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt &= (~(SAM|EDMA_FWID) );
	    /*set SAM and program FWID*/
	    ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt |= \
		(mode |((width & 0x7) << 8));
	}
    }
}


/******************************************************************************
 *
 * DMA destination parameters setup
 * ARGUMENTS:
 *    lch - channel or param device for destination parameters to be configured
 *    dest_port    - Destination port address
 *    addressMode  - indicates wether addressing mode is fifo.
 *
 *****************************************************************************/
void davinci_set_dma_dest_params(int lch, unsigned long dest_port, \
    enum address_mode mode, enum fifo_width width)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	/* set the destination port address in dest register of param structure */
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].dst = dest_port;

	/* set the fifo addressing mode */
	if (mode)
	{   /*reset DAM and FWID*/
	    ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt &= (~( DAM | EDMA_FWID));
	    /*set DAM and program FWID*/
	    ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt |= \
		((mode << 1) |((width & 0x7) << 8));
	}
    }
}


/******************************************************************************
 *
 * DMA source index setup
 * ARGUMENTS:
 *      lch     - channel or param device for configuration of source index
 *      srcbidx - source B-register index
 *      srccidx - source C-register index
 *
 *****************************************************************************/
void davinci_set_dma_src_index(int lch, short srcbidx, short srccidx)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_bidx &= 0xffff0000;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_bidx |= srcbidx;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_cidx &= 0xffff0000;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_cidx |= srccidx;
    }
}

/******************************************************************************
 *
 * DMA destination index setup
 * ARGUMENTS:
 *      lch    - channel or param device for configuration of destination index
 *      srcbidx - dest B-register index
 *      srccidx - dest C-register index
 *
 *****************************************************************************/
void davinci_set_dma_dest_index(int lch, short destbidx, short destcidx)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_bidx &= 0x0000ffff;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_bidx |= \
			  ((unsigned long)destbidx << 16);
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_cidx &= 0x0000ffff;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].src_dst_cidx |= \
	    ((unsigned long)destcidx << 16);
    }
}

/******************************************************************************
 *
 * DMA transfer parameters setup
 * ARGUMENTS:
 *      lch  - channel or param device for configuration of aCount, bCount and
 *         cCount regs.
 *      acnt - acnt register value to be configured
 *      bcnt - bcnt register value to be configured
 *      ccnt - ccnt register value to be configured
 *
 *****************************************************************************/
void davinci_set_dma_transfer_params(int lch, unsigned short acnt, \
    unsigned short bcnt, unsigned short ccnt, unsigned short bcntrld, \
    enum sync_dimension sync_mode)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].link_bcntrld &= 0x0000ffff ;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].link_bcntrld |= (bcntrld<<16) ;
	if(sync_mode == Async)
	    ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt &= (~SYNCDIM);
	else
	    ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt |= SYNCDIM;

		/* Set the acount, bcount, ccount registers */
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].a_b_cnt = (bcnt << 16) | acnt;
//        ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].ccnt &= 0xffff0000;
//        ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].ccnt |= ccnt;
	ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].ccnt = ccnt;

    }
}

/******************************************************************************
 *
 * davinci_set_dma_params -
 * ARGUMENTS:
 *      lch - logical channel number
 *
 *****************************************************************************/
void davinci_set_dma_params(int lch, edmacc_paramentry_regs *temp)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{
		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	memcpylocal((void *)&(ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt), (void *)temp, \
	sizeof(edmacc_paramentry_regs));
    }
}

/******************************************************************************
 *
 * davinci_get_dma_params -
 * ARGUMENTS:
 *      lch - logical channel number
 *
 *****************************************************************************/
void davinci_get_dma_params(int lch, edmacc_paramentry_regs *temp)
{
	int temp_ch=0;
	if (lch >= DAVINCI_EDMA_NUM_DMACH && lch < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch-DAVINCI_EDMA_NUM_DMACH];
		lch = temp_ch;
	}

    if(lch >=0 && lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {
	memcpylocal((void *)temp ,(void *)&(ptr_edmacc_regs->paramentry[dma_chan[lch].param_no].opt) , \
	sizeof(edmacc_paramentry_regs));
    }
}


/******************************************************************************
 *
 * DMA Strat - Starts the dma on the channel passed
 * ARGUMENTS:
 *      lch - logical channel number
 *
 *****************************************************************************/
int davinci_start_dma(int lch)
{
    int retval;

    if (lch >=0 && (lch < DAVINCI_EDMA_NUM_DMACH ))
    {
	int i = 0;
	int flag = 0;

	/* If the dma start request is for the unused events */
	while (dmachan_noevent[i] != -1)
	{
	    if (dmachan_noevent[i] == lch)
	    {
		/*EDMA channels without event association*/
//#ifdef EDMA_DEBUG
		DEBUGMSG("ESR=%x\r\n",ptr_edmacc_regs->shadow[0].esr);
//#endif

		(lch < 32)?(ptr_edmacc_regs->shadow[0].esr |= (1UL << lch)): \
		    (ptr_edmacc_regs->shadow[0].esrh |= (1UL << (lch-32)));


		flag = 1;
		retval = 0;
		break;
	    }
	    i++;
	}

	if (!flag)
	{
	    /*EDMA channel with event association*/
#ifdef EDMA_DEBUG
	    DEBUGMSG( "ER=%d\r\n",ptr_edmacc_regs->shadow[0].er);
#endif
			/*Clear any pedning error*/
	    (lch < 32)?(ptr_edmacc_regs->emcr |= (1UL << lch)): \
		(ptr_edmacc_regs->emcrh |= (1UL << (lch-32)));

			/*Clear any SER*/
	    (lch < 32)?(ptr_edmacc_regs->shadow[0].secr |= (1UL << lch)): \
		(ptr_edmacc_regs->shadow[0].secrh |= (1UL << (lch-32)));

	    (lch < 32)?(ptr_edmacc_regs->shadow[0].eesr |= (1UL << lch)): \
		(ptr_edmacc_regs->shadow[0].eesrh |= (1UL << (lch-32)));
#ifdef EDMA_DEBUG
	    DEBUGMSG("EER=%d\r\n",ptr_edmacc_regs->shadow[0].eer);
#endif
	    retval = 0;

	}
    }
    else if((lch >= DAVINCI_EDMA_NUM_DMACH) && (lch < (DAVINCI_EDMA_NUM_DMACH+ \
	DAVINCI_EDMA_NUM_QDMACH)))
    {
	ptr_edmacc_regs->shadow[0].qeesr |= (1<<(lch-DAVINCI_EDMA_NUM_DMACH));

	retval = 0;
    }
    else
    {   /* for slaveChannels */
	retval = EINVAL;
    }

    return retval;
}


/******************************************************************************
 *
 * DMA Stop - Stops the dma on the channel passed
 * ARGUMENTS:
 *      lch - logical channel number
 *
 *****************************************************************************/
void davinci_stop_dma(int lch)
{
    if (lch < DAVINCI_EDMA_NUM_DMACH)
    {
	int flag = 0;
	int i = 0;

	/* If the dma stop request is for the unused events */
	while (dmachan_noevent[i] != -1)
	{
	    if (dmachan_noevent[i] == lch)
	    {
		/*EDMA channels without event association*/

		/* if the requested channel is one of the unused channels the
		   reset the coresponding bit of ESR-Event Set Register */
		//ptr_edmacc_regs->ESR =
		//ptr_edmacc_regs->ESR | (0 << lch);
		flag = 1;
		break;
	    }
	    i++;
	}

	if (!flag)
	{
	     /*EDMA channel with event association*/
	     (lch < 32)?(ptr_edmacc_regs->shadow[0].eecr |= (1UL << lch)): \
		(ptr_edmacc_regs->shadow[0].eecrh |= (1UL << (lch-32)));


	     if(lch<32)
	     {
		 if(ptr_edmacc_regs->shadow[0].er & (1<<lch))
		 {
#ifdef EDMA_DEBUG
		     DEBUGMSG("ER=%x\n",ptr_edmacc_regs->shadow[0].er);
#endif
		     ptr_edmacc_regs->shadow[0].ecr |= (1<<lch);

		 }
	     }
	     else
	     {

		 if(ptr_edmacc_regs->shadow[0].erh & (1<<(lch-32)))
		 {
#ifdef EDMA_DEBUG
		     DEBUGMSG("ERH=%x\n",ptr_edmacc_regs->shadow[0].erh);
#endif
		     ptr_edmacc_regs->shadow[0].ecrh |= (1<<(lch-32));

		 }
	     }

	     if(lch<32)
	     {

		 if(ptr_edmacc_regs->shadow[0].ser & (1<<lch))
		 {
#ifdef EDMA_DEBUG
		     DEBUGMSG("SER=%x\n",ptr_edmacc_regs->shadow[0].ser);
#endif
		     ptr_edmacc_regs->shadow[0].secr |= (1<<lch);

		 }
	     }
	     else
	     {

		 if(ptr_edmacc_regs->shadow[0].serh & (1<<(lch-32)))
		 {
		     DEBUGMSG("SERH=%x\n",ptr_edmacc_regs->shadow[0].serh);
		     ptr_edmacc_regs->shadow[0].secrh |= (1<<(lch-32));

		 }
	     }

	     if(lch<32)
	     {

		 if(ptr_edmacc_regs->emr & (1<<lch))
		 {
#ifdef EDMA_DEBUG
		     DEBUGMSG("EMR=%x\n",ptr_edmacc_regs->emr);
#endif
		     ptr_edmacc_regs->emcr |= (1<<lch);

		 }
	     }
	     else
	     {

		 if(ptr_edmacc_regs->emrh & (1<<(lch-32)))
		 {
		     DEBUGMSG("EMRH=%x\n",ptr_edmacc_regs->emrh);
		     ptr_edmacc_regs->emcrh |= (1<<(lch-32));

		 }
	     }

#ifdef EDMA_DEBUG
		DEBUGMSG("EER=%d\r\n",ptr_edmacc_regs->shadow[0].eer);
#endif
	    /* if the requested channel is one of the event channels then just
	      set the link field of the corresponding param entry to 0xffff */
//            ptr_edmacc_regs->paramentry[lch].link_bcntrld &=   0xffff0000;
//            ptr_edmacc_regs->paramentry[lch].link_bcntrld |= 0xffff;
	}
    }
    else if((lch >= DAVINCI_EDMA_NUM_DMACH) && \
	(lch < (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH)))
    {
	/*for QDMA channels*/
	ptr_edmacc_regs->qeecr |= (1<<(lch-DAVINCI_EDMA_NUM_DMACH));
#ifdef EDMA_DEBUG
	DEBUGMSG("QER=%d\r\n",ptr_edmacc_regs->qer);
	DEBUGMSG("QEER=%d\r\n",ptr_edmacc_regs->qeer);
#endif
    }
    else if((lch >= (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH)) && \
	lch < DAVINCI_EDMA_NUM_PARAMENTRY)
    {   /* for slaveChannels */
	ptr_edmacc_regs->paramentry[lch].link_bcntrld &= 0xffff0000;
	ptr_edmacc_regs->paramentry[lch].link_bcntrld |= 0xffff;

    }
    else
    {}
}


/******************************************************************************
 *
 * DMA channel link - link the two logical channels passed through by linking
 *                    the link field of head to the param pointed by the lch_queue.
 * ARGUMENTS:
 *      lch_head  - logical channel number, in which the link field is linked
 *                  to the param pointed to by lch_queue
 * lch_queue - logical channel number or the param entry number, which is to be
 *                  linked to the lch_head
 *
 *****************************************************************************/
void davinci_dma_link_lch(int lch_head, int lch_queue)
{
    unsigned long link;
    int temp_ch=0;

    if (lch_head >= DAVINCI_EDMA_NUM_DMACH && lch_head < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	temp_ch = qdam_to_param_mapping[lch_head-DAVINCI_EDMA_NUM_DMACH];
	lch_head = temp_ch;
    }

    if (lch_queue >= DAVINCI_EDMA_NUM_DMACH && lch_queue < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	temp_ch = qdam_to_param_mapping[lch_queue-DAVINCI_EDMA_NUM_DMACH];
	lch_queue = temp_ch;
    }

    if ((lch_head >= 0 && lch_head < DAVINCI_EDMA_NUM_PARAMENTRY) &&  \
	(lch_queue >= 0 && lch_queue < DAVINCI_EDMA_NUM_PARAMENTRY))
    {
	/*program LINK*/
	link=(unsigned long)(&(ptr_edmacc_regs->paramentry[dma_chan[lch_queue].param_no].opt));
	ptr_edmacc_regs->paramentry[dma_chan[lch_head].param_no].link_bcntrld &= 0xffff0000;
	ptr_edmacc_regs->paramentry[dma_chan[lch_head].param_no].link_bcntrld |= \
	    ((unsigned short)link);
	dma_chan[lch_head].link_lch = lch_queue;
    }
}


/******************************************************************************
 *
 * DMA channel unlink - unlink the two logical channels passed through by
 *                   setting the link field of head to 0xffff.
 * ARGUMENTS:
 * lch_head - logical channel number, from which the link field is to be removed
 * lch_queue - logical channel number or the param entry number, which is to be
 *             unlinked from lch_head
 *
 *****************************************************************************/
void davinci_dma_unlink_lch(int lch_head, int lch_queue)
{
    int temp_ch=0;

    if (lch_head >= DAVINCI_EDMA_NUM_DMACH && lch_head < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	temp_ch = qdam_to_param_mapping[lch_head-DAVINCI_EDMA_NUM_DMACH];
	lch_head = temp_ch;
    }

    if (lch_queue >= DAVINCI_EDMA_NUM_DMACH && lch_queue < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	temp_ch = qdam_to_param_mapping[lch_queue-DAVINCI_EDMA_NUM_DMACH];
	lch_queue = temp_ch;
    }

    if ((lch_head >= 0 && lch_head < DAVINCI_EDMA_NUM_PARAMENTRY) &&  \
	(lch_queue >= 0 && lch_queue < DAVINCI_EDMA_NUM_PARAMENTRY))
    {
	ptr_edmacc_regs->paramentry[dma_chan[lch_head].param_no].link_bcntrld |= 0xffff;
	dma_chan[lch_head].link_lch = -1;
    }
}

/******************************************************************************
 *
 * DMA channel chain - chains the two logical channels passed through by
 * ARGUMENTS:
 * lch_head - logical channel number, from which the link field is to be removed
 * lch_queue - logical channel number or the param entry number, which is to be
 *             unlinked from lch_head
 *
 *****************************************************************************/
void davinci_dma_chain_lch(int lch_head, int lch_queue)
{

	int temp_ch=0;
	if (lch_head >= DAVINCI_EDMA_NUM_DMACH && lch_head < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch_head-DAVINCI_EDMA_NUM_DMACH];
		lch_head = temp_ch;
	}

	if (lch_queue >= DAVINCI_EDMA_NUM_DMACH && lch_queue < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
	{

		temp_ch = qdam_to_param_mapping[lch_queue-DAVINCI_EDMA_NUM_DMACH];
		lch_queue = temp_ch;
	}

    if ((lch_head >= 0 && \
	lch_head < (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH)) &&  \
	(lch_queue >= 0 && \
	lch_queue < (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH)))
    {/*set TCCHEN*/
     //ptr_edmacc_regs->paramentry[lch_head].opt &= ~(TCCHEN); /*This configuration is not working*/
     /*set ITCCHEN*/
	ptr_edmacc_regs->paramentry[lch_head].opt |= ITCCHEN;

    /*program tcc*/
	ptr_edmacc_regs->paramentry[lch_head].opt &= (~TCC);
	ptr_edmacc_regs->paramentry[lch_head].opt |= (lch_queue & 0x3f) << 12;
    }
}

/******************************************************************************
 *
 * DMA channel unchain - unchain the two logical channels passed through by
 * ARGUMENTS:
 * lch_head - logical channel number, from which the link field is to be removed
 * lch_queue - logical channel number or the param entry number, which is to be
 *             unlinked from lch_head
 *
 *****************************************************************************/
void davinci_dma_unchain_lch(int lch_head, int lch_queue)
{
    int temp_ch=0;

    if (lch_head >= DAVINCI_EDMA_NUM_DMACH && lch_head < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	temp_ch = qdam_to_param_mapping[lch_head-DAVINCI_EDMA_NUM_DMACH];
	lch_head = temp_ch;
    }

    if (lch_queue >= DAVINCI_EDMA_NUM_DMACH && lch_queue < \
       (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH))
    {
	temp_ch = qdam_to_param_mapping[lch_queue-DAVINCI_EDMA_NUM_DMACH];
	lch_queue = temp_ch;
    }

    if ((lch_head >= 0 && \
	lch_head < (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH)) &&  \
	(lch_queue >= 0 && \
	lch_queue < (DAVINCI_EDMA_NUM_DMACH + DAVINCI_EDMA_NUM_QDMACH)))
    {
	/*reset TCCHEN*/
	//ptr_edmacc_regs->paramentry[lch_head].OPT &= ~TCCHEN;
	/*reset ITCCHEN*/
	ptr_edmacc_regs->paramentry[lch_head].opt &= (~ITCCHEN);
    }
}




/******************************************************************************
 *
 * It cleans ParamEntry qand bring back EDMA to initial state if media has
 * been removed before EDMA has finished.It is usedful for removable media.
 * Arguments:
 *      chno     - channel no
 *
 * Return: zero on success, or corresponding error no on failure
 *
 *****************************************************************************/

void davinci_clean_channel(int chno)
{
    int i;


#ifdef EDMA_DEBUG
    DEBUGMSG("EMR =%d\r\n",ptr_edmacc_regs->emr);
#endif
    if(chno<32)
    {
	for(i=0; i<32;i++)
	{
	    if(chno ==i)
	    {
		ptr_edmacc_regs->shadow[0].ecr |= (1<<i);
		/* here write to EMCR to clear the corresponding EMR bits*/
		ptr_edmacc_regs->emcr |= (1<<i);
		/*Clear any SER*/
		ptr_edmacc_regs->shadow[0].secr |= (1<<i);
		ptr_edmacc_regs->ccerrclr |=((1<<16)|0x3);

	    }
	}
    }

    if(chno>32)
    {
#ifdef EDMA_DEBUG
	DEBUGMSG("EMRH =%d\r\n",ptr_edmacc_regs->emrh);
#endif
	for(i=0; i<32;i++)
	{
	    if(chno == (i+32))
	    {
		ptr_edmacc_regs->shadow[0].ecrh |= (1<<i);
		/* here write to ICR to clear the corresponding IPR bits*/
		ptr_edmacc_regs->emcrh |= (1<<i);
		/*Clear any SER*/
		ptr_edmacc_regs->shadow[0].secrh |= (1<<i);
		ptr_edmacc_regs->ccerrclr |=((1<<16)|0x3);
	    }
	}
    }
}
arch_initcall(arch_dma_init);

EXPORT_SYMBOL(davinci_request_dma);
EXPORT_SYMBOL(davinci_free_dma);
EXPORT_SYMBOL(davinci_set_dma_src_params);
EXPORT_SYMBOL(davinci_set_dma_dest_params);
EXPORT_SYMBOL(davinci_set_dma_src_index);
EXPORT_SYMBOL(davinci_set_dma_dest_index);
EXPORT_SYMBOL(davinci_set_dma_transfer_params);
EXPORT_SYMBOL(davinci_set_dma_params);
EXPORT_SYMBOL(davinci_get_dma_params);
EXPORT_SYMBOL(davinci_start_dma);
EXPORT_SYMBOL(davinci_stop_dma);
EXPORT_SYMBOL(davinci_dma_link_lch);
EXPORT_SYMBOL(davinci_dma_unlink_lch);
EXPORT_SYMBOL(davinci_dma_chain_lch);
EXPORT_SYMBOL(davinci_dma_unchain_lch);
EXPORT_SYMBOL(davinci_clean_channel);
