/*
 * linux/arch/arm/mach-davinci/dma_osdep.c
 *
 * TI DaVinci DMA registration file
 *
 * Copyright (C) 2004 Texas Instruments.
 *
 * ----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------------
 Modifications:
 ver. 1.0: Oct 2005, Swaminathan S
 -
 *
 */

#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/arch/memory.h>
#include <asm/arch/edma.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>

#define DAVINCI_DMA_REGISTER_BASE DAVINCI_DMA_3PCC_BASE

typedef void (*intr_callback) (void);

edmacc_regs *get_edma_base(void)
{
     return ((edmacc_regs *)IO_ADDRESS(DAVINCI_DMA_REGISTER_BASE));
}

static intr_callback cb[4];

/******************************************************************************
 *
 * DMA interrupt handlers
 *
 *****************************************************************************/
static int dma_irq_handler(int sound_curr_lch, void *ch_status,
			   struct pt_regs *data)
{
	//printk(KERN_INFO"dma_irq_handler\n");
	(*cb[0]) ();
	return IRQ_HANDLED;
}

static int dma_ccerr_handler(int sound_curr_lch, void *ch_status,
			     struct pt_regs *data)
{
	printk(KERN_INFO "dma_ccerr_handler\n");
	(*cb[1]) ();
	return IRQ_HANDLED;
}

static int dma_tc1err_handler(int sound_curr_lch, void *ch_status,
			      struct pt_regs *data)
{
	printk(KERN_INFO "dma_tc1err_handler\n");
	(*cb[2]) ();
	return IRQ_HANDLED;
}

static int dma_tc2err_handler(int sound_curr_lch, void *ch_status,
			      struct pt_regs *data)
{
	printk(KERN_INFO "dma_tc2err_handler\n");
	(*cb[3]) ();
	return IRQ_HANDLED;
}

int register_dma_interrupts(intr_callback cb1, intr_callback cb2,
			    intr_callback cb3, intr_callback cb4)
{
	cb[0] = cb1;
	cb[1] = cb2;
	cb[2] = cb3;
	cb[3] = cb4;
	if (!cb1 || !cb2 || !cb3 || !cb4) {
		printk(KERN_INFO "NULL callback\n");
		return -1;
	}

	if(request_irq(IRQ_CCINT0,dma_irq_handler,0,"EDMA",NULL))
	{
		printk(KERN_INFO "request_irq failed\n");
		return -1;
	}
	if(request_irq(IRQ_CCERRINT ,dma_ccerr_handler,0,"EDMA CC Err",NULL))
	{
		printk(KERN_INFO "request_irq failed\n");
		return -1;
	}
	if(request_irq(IRQ_TCERRINT0,dma_tc1err_handler,0,"EDMA TC1 Err",NULL))
	{
		printk(KERN_INFO "request_irq failed\n");
		return -1;
	}
	if(request_irq(IRQ_TCERRINT,dma_tc2err_handler,0,"EDMA TC2 Err",NULL))
	{
		printk(KERN_INFO "request_irq failed\n");
		return -1;
	}
	return 0;
}
