/*
 * linux/arch/arm/mach-davinci/time.c
 *
 * TI DaVinci timer file
 *
 * Copyright (C) 2004 Texas Instruments.
 *
 * ----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------------
 Modifications:
 ver. 1.0: Oct 2005, Swaminathan S
 -
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/timex.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/interrupt.h>

#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/system.h>
#include <asm/leds.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>

#include <asm/arch/timex.h>
#include <asm/arch/irqs.h>

#define TIMER_BASE_ADDR IO_ADDRESS(DAVINCI_TIMER1_BASE)
#define IRQ_TIMER IRQ_TINT1_TINT34

#define TIMER_COUNTS_UP 1

static timer0_registers *davinci_timer0 = (timer0_registers *) TIMER_BASE_ADDR;

/*
 * Last processed system timer interrupt
 */
static unsigned long davinci_timer_last = 0;

unsigned long davinci_gettimeoffset(void);

static irqreturn_t davinci_timer0_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	davinci_timer_last = davinci_timer0->tim34;
	/* Do the Linux timer operations */
	timer_tick(regs);

	return IRQ_HANDLED;
}

static struct irqaction davinci_timer_irq = {
	.name = "Timer0",
	.flags = SA_INTERRUPT,
	.handler = davinci_timer0_interrupt
};

static void setup_timer(void)
{
	davinci_timer0->tcr = 0x00;	/* Timer12 disabled, Timer34 disabled */

	/* Set timers to unchained dual 32 bit timers, Unreset timer34, timer12 */
	davinci_timer0->tgcr = 0;
	davinci_timer0->tgcr = 0x7;

	/* Program the timer34 counter register - set the prd34 for right count */
	davinci_timer0->tim34 = 0;

	/* Program the timer12 counter register - set the prd12 for right count
	 * for DSP timer
	 */
	davinci_timer0->tim12 = 0;

	/* The timer is programmed to expire after "10 mSecs" milliseconds */
	davinci_timer0->prd34 = (CLOCK_SOURCE_FREQUENCY / (HZ));

	/* The timer is programmed to expire after "10 mSecs" milliseconds */
	davinci_timer0->prd12 = (CLOCK_SOURCE_FREQUENCY / (HZ));

	/* Register the time interrupt */
	setup_irq(IRQ_TIMER, &davinci_timer_irq);

	/* Start the timer */
	/* Enable timer12, timer34 (for DSP) */
	davinci_timer0->tcr = ((0x80 << 16) | (0x80));
	/* Timer34 continously enabled, Timer12 continuouzly enabled */

}

void __init davinci_timer_init(void)
{
	setup_timer();
}

unsigned long davinci_gettimeoffset(void)
{
	unsigned long elapsed, now;
	unsigned long oneusec = DAVINCI_TIMER_FREQUENCY / 1000000;

	disable_irq (IRQ_TIMER);
	now = davinci_timer0->tim34;

	if (davinci_timer_last != 0) {
		elapsed = now - davinci_timer_last;
	} else {
		elapsed = now;
	}

	davinci_timer_last = now;
	enable_irq (IRQ_TIMER);

	return (((elapsed / oneusec) == 0) ? 1 : (elapsed / oneusec));
}

struct sys_timer davinci_timer = {
	.init = davinci_timer_init,
	.offset = davinci_gettimeoffset,
};
