/*
 * $Id: dbox2-flash.c,v 1.13 2004/11/04 13:24:14 gleixner Exp $
 *
 * D-Box 2 flash driver
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/squashfs_fs.h>
#include <asm/mips-boards/prom.h>
#include <asm/addrspace.h>
#include <linux/squashfs_fs.h>

#define WINDOW_ADDR     CONFIG_MTD_PHYSMAP_START
#define FLASH_BASE      (CONFIG_MTD_PHYSMAP_START | 0xA0000000)
#define RAM_BASE        (0x94000000 | 0xA0000000)
#define WINDOW_SIZE     CONFIG_MTD_PHYSMAP_LEN

static struct mtd_info *mymtd;
static struct mtd_info *mymtdram;


struct map_info ohio_flash_map = {
	.name		= "Ohio flash memory",
	.size		= WINDOW_SIZE,
	.bankwidth	= 2,
	.phys		= WINDOW_ADDR,
};

struct map_info ohio_ram_map = {
	.name		= "Ohio virtual ROM memory",
	.size		= 0,
	.bankwidth	= 2,
	.phys		= 0,
};

extern int root_dev_setup(char *line);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int ohio_find_hidden_filesystem(unsigned int *size, unsigned int *start, unsigned int *fs_size, unsigned int ram) {
    unsigned int start_pos, first_start_pos;
    unsigned int end_pos;
    squashfs_super_block *super;

    if(ram) {
        start_pos  = (unsigned int)*start + RAM_BASE;
        end_pos   = (unsigned int)(*start + RAM_BASE + *size);
    } else {
        start_pos  = (unsigned int)*start + FLASH_BASE;
        end_pos   = (unsigned int)(*start + FLASH_BASE + *size);
    }
    /*--- printk("[ohio_find_hidden_filesystem] start=0x%x end=0x%x\n", start_pos, end_pos); ---*/

    first_start_pos = start_pos;
    while(start_pos < end_pos) {
        /*--- printk("[0x%x]", *((unsigned int *)start_pos)); ---*/
        if(*((unsigned int *)start_pos) == SQUASHFS_MAGIC) {
            super = (squashfs_super_block *)start_pos;
            printk("[ohio_find_hidden_filesystem]: super block found: bytes_used: 0x%x/%u\n",
                    super->bytes_used, super->bytes_used);
            if(fs_size)
                *fs_size = super->bytes_used + sizeof(*super) + (start_pos - first_start_pos);
            if(ram) {
                *start = start_pos - RAM_BASE;
            } else {
                *start = start_pos - FLASH_BASE;
            }
            *size  = end_pos - start_pos;
            return 0;
        }
        start_pos += 256;
    }
    return 1;
}   

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

#define LOCAL_MTD_FILESYSTEM                0
#define LOCAL_MTD_KERNEL                    1
#define LOCAL_MTD_URLADER                   2
#define LOCAL_MTD_TFFS1                     3
#define LOCAL_MTD_TFFS2                     4
#define LOCAL_URLADER_MTDS                  5

#define LOCAL_MTD_JFFS2                     5
#define LOCAL_MTD_KERNEL_WITHOUT_JFFS2      6
#define LOCAL_MAX_MTDS                      7

#define _NUM_PARTITIONS                     LOCAL_MAX_MTDS
unsigned int NUM_PARTITIONS  = _NUM_PARTITIONS;
static struct mtd_partition partition_info[LOCAL_MAX_MTDS];
static struct mtd_partition partition_info_ram[2];

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
char *get_mtdram1_param(void) {
    char *cmdline = prom_getcmdline();
    return strstr(cmdline, "mtdram1");
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int __init init_ohio_flash(void) {
    unsigned int flash_size = 0;
    unsigned int mtd_base[LOCAL_MAX_MTDS];
    unsigned int mtd_size[LOCAL_MAX_MTDS];
    unsigned int mtd_index;
    char *p;
    char name[10];
    unsigned int flash_mask;
    unsigned int jffs2_size = 0;

    printk(KERN_NOTICE "Ohio flash driver (size->0x%X mem->0x%X)\n", WINDOW_SIZE, WINDOW_ADDR);
    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
    p = prom_getenv("flashsize");
    if(!p) {
        printk(KERN_ERR "[init_ohio_flash] %s not found\n", "flashsize");
	    return -ENXIO;
    }
    flash_size = simple_strtol(p, NULL, 0);
    printk("flash_size=0x%x\n", flash_size);
    ohio_flash_map.size = flash_size;
    flash_mask = flash_size * 2 - 1;

    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
    partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].name = "Kernel without jffs2";
    partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].offset     = 0;
    partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].size       = 0;
    partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].mask_flags = 0;

    partition_info[LOCAL_MTD_JFFS2].name = "jffs2";
    partition_info[LOCAL_MTD_JFFS2].offset     = 0;
    partition_info[LOCAL_MTD_JFFS2].size       = 0;
    partition_info[LOCAL_MTD_JFFS2].mask_flags = 0;

    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
    for(mtd_index = 0 ; mtd_index < LOCAL_URLADER_MTDS ; mtd_index++) {
        char *p1;
        char Buffer[64];
        sprintf(name, "mtd%u", mtd_index);
        p = prom_getenv(name);
        if(!p) {
            printk(KERN_ERR "[init_ohio_flash] %s not found\n", name);
            return -ENXIO;
        }
#if 0
        printk("env=\"%s\"=\"%s\"\n", name, p);
#endif
        strcpy(Buffer, p);
        p1 = p = Buffer;
        while(*p1 && *p1 != ',') p1++;
        if(*p1) {
            *p1++ = '\0';
            /*----------------------------------------------------------------------------------*\
            \*------------------------------------------------------------------------------*/
            mtd_base[mtd_index] = CPHYSADDR(simple_strtol(p, NULL, 0)) & flash_mask;
            mtd_size[mtd_index] = (CPHYSADDR(simple_strtol(p1, NULL, 0)) & flash_mask) - mtd_base[mtd_index];
#if 0
            printk("[init_ohio_flash]: %s base=0x%x size=0x%x\n", name, mtd_base[mtd_index], mtd_size[mtd_index]);
#endif
            partition_info[mtd_index].size       = mtd_size[mtd_index];
            partition_info[mtd_index].offset     = mtd_base[mtd_index];
            switch(mtd_index) {
                case 2:
                    partition_info[mtd_index].name = "bootloader";
                    partition_info[mtd_index].mask_flags = 0;
                    break;
                case 1:
                    partition_info[mtd_index].name = "kernel";
                    partition_info[mtd_index].mask_flags = MTD_NO_VIRTBLOCKS;
                    break;
                case 0:
                    partition_info[mtd_index].name = "filesystem";
                    partition_info[mtd_index].mask_flags = MTD_NO_VIRTBLOCKS;
                    break;
                case 3:
                    partition_info[mtd_index].name = "tffs (1)";
                    partition_info[mtd_index].mask_flags = 0;
                    break;
                case 4:
                    partition_info[mtd_index].name = "tffs (2)";
                    partition_info[mtd_index].mask_flags = 0;
                    break;
            }
            if(partition_info[mtd_index].size == 0)
                partition_info[mtd_index].offset = 0;

#if 0
            printk("partition_info[%u]: name=%s offset=%x size=%x\n",
                    mtd_index, 
                    partition_info[mtd_index].name,
                    partition_info[mtd_index].offset,
                    partition_info[mtd_index].size);
            printk("partition_info[%u]: name=%s offset=%x size=%x\n",
                    LOCAL_MAX_MTDS + 0, 
                    partition_info[LOCAL_MAX_MTDS + 0].name,
                    partition_info[LOCAL_MAX_MTDS + 0].offset,
                    partition_info[LOCAL_MAX_MTDS + 0].size);
            printk("partition_info[%u]: name=%s offset=%x size=%x\n",
                    LOCAL_MAX_MTDS + 1, 
                    partition_info[LOCAL_MAX_MTDS + 1].name,
                    partition_info[LOCAL_MAX_MTDS + 1].offset,
                    partition_info[LOCAL_MAX_MTDS + 1].size);
#endif
        }
    }

    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
	ohio_flash_map.virt = ioremap(WINDOW_ADDR, flash_size);

	if (!ohio_flash_map.virt) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	simple_map_init(&ohio_flash_map);

	// Probe for dual Intel 28F320 or dual AMD
	mymtd = do_map_probe("cfi_probe", &ohio_flash_map);
	if (!mymtd) {
	    // Probe for single Intel 28F640
	    ohio_flash_map.bankwidth = 2;
	
	    mymtd = do_map_probe("cfi_probe", &ohio_flash_map);
	}
	    
    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
	if (mymtd) {
        signed int i;
        unsigned int hidden_size = 0, hidden_start = 0;
        unsigned int fs_size = 0;
		mymtd->owner = THIS_MODULE;

        for(i = LOCAL_URLADER_MTDS - 1 ; i >= 0 ; i--) {
            /*--- printk("partition_info[i].name = \"%s\"\n", partition_info[i].name); ---*/
            if(!strcmp(partition_info[i].name, "kernel")) {
                char *jffs2_ptr = prom_getenv("jffs2_size");
                if(jffs2_ptr) {
                    jffs2_size = simple_strtol(jffs2_ptr, NULL, 0) * (1 << 16);
                    printk("[mtd]: jffs2_size = %s * 64KByte (0x%x Bytes)\n", jffs2_ptr, jffs2_size);
                } else {
                    jffs2_size = 2 * (1 << 20);
                    printk("[mtd]: set to default: jffs2_size = 0x%x * 64KByte (0x%x Bytes)\n", jffs2_size / (1 << 16), jffs2_size);
                }

                hidden_size = partition_info[i].size, hidden_start = partition_info[i].offset;

                if(ohio_find_hidden_filesystem(&hidden_size, &hidden_start, &fs_size, 0)) {
                    printk("[init_ohio_flash] find hidden root filesystem failed\n");
	                iounmap((void *)ohio_flash_map.virt);
                    hidden_size = 0, hidden_start = 0;
                    if(get_mtdram1_param()) {
                        printk("[mtdram]: no filesystem found in FLASH, use hole partition for posible jffs2\n");
                        hidden_size = partition_info[i].size, hidden_start = partition_info[i].offset;
                        printk("[mtdram] filesystem size=0x%x offset=0x%x\n", hidden_size, hidden_start);
                        fs_size = 1; /*--- dummy: klein, nur nicht 0 ---*/
                    }
                } else {
                    printk("[init_ohio_flash] find hidden filesystem size=0x%x offset=0x%x\n", hidden_size, hidden_start);
                }
                if(jffs2_size && fs_size) {
                    printk("[mtd] configure jffs2 partition\n");
                    fs_size |= (1 << 16) - 1; fs_size += 1;  /* aufrunden auf naechstes 64K segment */
                    printk("[mtd] fs_size=0x%x max=0x%x is=0x%x max jffs2_size value %u\n", 
                            fs_size, partition_info[i].size - fs_size, 
                            jffs2_size, (partition_info[i].size - fs_size) / (1 << 16));

                    if(partition_info[i].size - fs_size < jffs2_size) {
                        printk("[mtd] ERROR jffs2_size too big, reduce to max\n");
                        jffs2_size = partition_info[i].size - fs_size;
                    }

                    /*--- kernel without jffs ---*/
                    partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].size   = partition_info[i].size - jffs2_size;
                    partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].offset = partition_info[i].offset;

                    /*--- jffs ---*/
                    partition_info[LOCAL_MTD_JFFS2].size   = jffs2_size;
                    partition_info[LOCAL_MTD_JFFS2].offset = partition_info[i].offset + partition_info[LOCAL_MTD_KERNEL_WITHOUT_JFFS2].size;
                }
                continue;
            }
            if(!strcmp(partition_info[i].name, "filesystem")) {
                char buff[64];
                char *p;
                if((partition_info[i].size == 0) && (partition_info[i].offset == 0)) {
                    partition_info[i].size   = hidden_size;
                    partition_info[i].offset = hidden_start;
                }
                p = prom_getenv("nfsroot");
                if(p) {
                    root_dev_setup(p);
                } else {
                    sprintf(buff, "/dev/mtdblock%u", i + 1);
                    root_dev_setup(buff);
                }
            }
        }
        if(jffs2_size == 0)
            NUM_PARTITIONS -= 2;

        /*--- printk("[mtd] ohio_flash_map: name=%s bankwidth=%u virt=0x%p phys=0x%x size=0x%lx\n", ---*/
            /*--- ohio_flash_map.name, ohio_flash_map.bankwidth, ohio_flash_map.virt, ohio_flash_map.phys, ohio_flash_map.size); ---*/

        /* Create MTD devices for each partition. */
        mymtd->flags = MTD_CLEAR_BITS | MTD_CAP_NORFLASH | MTD_NO_VIRTBLOCKS;
        add_mtd_partitions(mymtd, partition_info, NUM_PARTITIONS);

        /*----------------------------------------------------------------------------------*\
         * MTDRAM
        \*----------------------------------------------------------------------------------*/
        {
            unsigned int mtdram_base, mtdram_size, ram_size, ram_mask;
            char *mtdram, *p1;
            mtdram = get_mtdram1_param();
            if(mtdram == NULL)
                return 0; /*--- der "normalfall" ---*/

            /*------------------------------------------------------------------------------*\
             * Aus der memsize die RAM-Addr-Maske ermitteln
            \*------------------------------------------------------------------------------*/
            p = prom_getenv("memsize");
            if(!p) {
	            return 0;
            }
            ram_size = simple_strtol(p, NULL, 0);
            printk("mem_size=0x%x\n", ram_size);
            ohio_ram_map.size = ram_size;
            ram_mask = ram_size & (((1 << 20) *64) - 1);  /* funktioniert bis 64 MByte Speicher */

            p1 = p = mtdram + sizeof("mtdram0=") - 1;
            while(*p1 && *p1 != ',') p1++;
            if(*p1 == '\0') {
                return 0;
            }

            *p1++ = '\0';
            mtdram_base = CPHYSADDR(simple_strtol(p, NULL, 0));
            printk("[mtdram] mtdram_base=0x%x ", mtdram_base);
            mtdram_size = CPHYSADDR(simple_strtol(p1, NULL, 0));
            printk(" mtdram_top=0x%x ", mtdram_size);
            mtdram_size -= mtdram_base;
            printk(" mtdram_size=0x%x\n", mtdram_size);

            /*------------------------------------------------------------------------------*\
            \*------------------------------------------------------------------------------*/
            printk("[mtdram]: ioremap(%x, %x)\n", mtdram_base, mtdram_size); 
            ohio_ram_map.virt = CKSEG0ADDR(ioremap(mtdram_base, mtdram_size));

            printk("[mtdram]: simple_map_init: ram mtd base is: 0x%x\n", mtdram_base);
	        simple_map_init(&ohio_ram_map);

            ohio_ram_map.size = mtdram_size;
            ohio_ram_map.phys = CPHYSADDR(mtdram_base);

            mtdram_base &= ram_mask;

            printk("[mtdram] ohio_ram_map: name=%s bankwidth=%u virt=0x%p phys=0x%x size=0x%lx\n",
                ohio_ram_map.name, ohio_ram_map.bankwidth, ohio_ram_map.virt, ohio_ram_map.phys, ohio_ram_map.size);

            printk("[mtdram]: do_map_probe\n");
            mymtdram = do_map_probe("map_ram", &ohio_ram_map);
            if(mymtdram) {
                printk("[mtdram]: do_map_probe success\n");
                mymtdram->owner = THIS_MODULE;
                mymtdram->flags = MTD_CAP_RAM;
            }

            partition_info_ram[0].name       = "filesystem in ram";
            partition_info_ram[0].mask_flags = 0;
            partition_info_ram[0].offset     = mtdram_base;   /* 0 ==> filesystem */
            partition_info_ram[0].size       = mtdram_size;
            partition_info_ram[1].name       = "kernel in ram";
            partition_info_ram[1].mask_flags = 0;
            partition_info_ram[1].offset     = mtdram_base;   /* kernel + filesystem ==> kernel */
            partition_info_ram[1].size       = mtdram_size;

            printk("[mtdram]: [0].offset=0x%x [0].size=0x%x [1].offset=0x%x [1].size=0x%x\n",
                partition_info_ram[0].offset, partition_info_ram[0].size, partition_info_ram[1].offset, partition_info_ram[1].size);

            if(ohio_find_hidden_filesystem(&(partition_info_ram[0].size), &(partition_info_ram[0].offset), NULL, 1)) {
                printk("[mtdram]: no hidden filesystem found\n");
	            iounmap((void *)ohio_ram_map.virt);
                return 0;
            }
            partition_info_ram[1].size   -= partition_info_ram[0].size;
            partition_info_ram[0].offset -= mtdram_base;
            partition_info_ram[1].offset -= mtdram_base;

            printk("[mtdram] ohio_ram_map: name=%s bankwidth=%u virt=0x%p phys=0x%x size=0x%lx\n",
                ohio_ram_map.name, ohio_ram_map.bankwidth, ohio_ram_map.virt, ohio_ram_map.phys, ohio_ram_map.size);
            printk("[mtdram]: [0].offset=0x%x [0].size=0x%x [1].offset=0x%x [1].size=0x%x\n",
                partition_info_ram[0].offset, partition_info_ram[0].size, partition_info_ram[1].offset, partition_info_ram[1].size);

            add_mtd_partitions(mymtdram, partition_info_ram, 2);
            {
                char buff[64];
                sprintf(buff, "/dev/mtdblock%u", NUM_PARTITIONS + 1);
                root_dev_setup(buff);
            }
        }
		return 0;
	}

	iounmap((void *)ohio_flash_map.virt);
	return -ENXIO;
}

static void __exit cleanup_ohio_flash(void)
{
	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (ohio_flash_map.virt) {
		iounmap((void *)ohio_flash_map.virt);
		ohio_flash_map.virt = 0;
	}
}

module_init(init_ohio_flash);
module_exit(cleanup_ohio_flash);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Martin Pommerenke m.pommerenke@avm.de");
MODULE_DESCRIPTION("MTD map driver for ohio board (copied from dbox2 files)");
