/*------------------------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*------------------------------------------------------------------------------------------*/
#include <stdarg.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <asm/io.h>
#if !defined(CONFIG_NETCHIP_ADM69961)
#define CONFIG_NETCHIP_ADM69961
#endif
#include <linux/avm_event.h>
#include "cpmac_if.h"
#include "cpmac_const.h"
#include "cpmac_types.h"
#include "cpmac_debug.h"
#include "cpmac_main.h"
#include "cpmac_sched.h"
#include "cpphy_const.h"
#include "cpphy_types.h"

extern dev_desc_t g_dev_array[CPMAC_MAX_PHY];
extern unsigned int cpmac_devices_installed;

#ifdef CONFIG_IP_MULTICAST_FASTFORWARD
extern struct mcfw_netdriver *cpmac_mcfw_netdrv;
extern int cpmac_mcfw_sourceid;
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static volatile void (*cpmac_wan_receive)(struct sk_buff *skb) = 0;

void cpmac_register_wan_receive(void (*wan_receive)(struct sk_buff *skb))
{
   local_bh_disable();
   cpmac_wan_receive = wan_receive;
   local_bh_enable();
}
EXPORT_SYMBOL(cpmac_register_wan_receive);

void cpmac_unregister_wan_receive(void)
{
   local_bh_disable();
   cpmac_wan_receive = 0;
   local_bh_enable();
}
EXPORT_SYMBOL(cpmac_unregister_wan_receive);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline cpmac_err_t cpmac_if_check_mac_handle (cpmac_priv_t *cpmac_priv)
{
  cpmac_err_t ret = CPMAC_ERR_NOERR;

  if (!cpmac_priv || !g_dev_array[cpmac_priv->inst].p_dev) {
    ret = CPMAC_ERR_ILL_HANDLE;
  }
  return ret;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_if_control_ind (cpmac_priv_t *cpmac_priv, cpmac_control_ind_t control, ...)
{
  va_list param;
  cpmac_err_t ret = CPMAC_ERR_NOERR;

  if (cpmac_if_check_mac_handle (cpmac_priv) != CPMAC_ERR_NOERR) {
    DEB_ERR("cpmac_if_control_ind, ill mac_handle %p\n",
        cpmac_priv);
  } else {
    DEB_INFO("cpmac_if_control_ind, got %u\n", control);
    switch (control) {
      case CPMAC_CONTROL_IND_MDIO_STATUS:
        {
          unsigned int linked;

          va_start (param, control);
          linked = va_arg(param, unsigned int); 
          cpmac_main_handle_mdio_status_ind (cpmac_priv, linked);
          va_end (param);
        }
        break;
      case CPMAC_CONTROL_IND_EVENT_UPDATE:
        cpmac_main_handle_event_update (cpmac_priv);
        break;
      case CPMAC_CONTROL_IND_GET_INSTANCE_COUNT:
        ret = (cpmac_err_t)cpmac_devices_installed;
        break;
      default:
        DEB_WARN("cpmac_if_control_ind, unknown control %u\n",
          control);
        break;
    }
  }
  return ret;
}
/*--- EXPORT_SYMBOL(cpmac_if_control_ind); ---*/

#ifdef CPMAC_DMA_RX_PRIOQUEUE_DEBUG
static char buf[60];
static int bufidx = 0;

static void log_from_dmaqueue(int wan)
{
  if (bufidx == sizeof(buf)-1) {
	 buf[bufidx] = 0;
	 DEB_INFO("cpmac: rx %s\n", buf);
	 bufidx = 0;
  }
  buf[bufidx++] = '0'+ (wan ? 1 : 0);
}
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_if_data_from_phy (cpmac_priv_t *cpmac_priv, struct sk_buff *skb, unsigned int length)
{
  cpmac_err_t ret = CPMAC_ERR_NOERR;
  int is_wan = 0;

  if (cpmac_if_check_mac_handle (cpmac_priv) != CPMAC_ERR_NOERR) {
    DEB_ERR("cpmac_if_data_from_phy, ill mac_handle %p\n",
        cpmac_priv);
    ret = CPMAC_ERR_ILL_HANDLE;
  } else {
    struct net_device *p_dev = cpmac_priv->owner;

    DEB_TRC("cpmac_if_data_from_phy, rcv %u bytes\n",
        length);
    __skb_trim (skb, length);

    /* invalidate the cache: data written by dma */
    dma_cache_inv((unsigned long)skb->data, skb->len);    

    /* 802.1q vlan stuff, discard received packet if: */
    /* port is not setup for VLAN but frame is vlan tagged or */
    /* is not vlan tagged but exceeds 1518 bytes or */
    /* port is setup for VLAN but frame is not vlan tagged and exceeds 1518 bytes */
    if ((!cpmac_priv->enable_vlan &&
        ((length > CPMAC_VLAN_TCI_END_OFFSET &&
        CPMAC_VLAN_IS_802_1Q_FRAME (skb->data)) ||
        (length > 1518))) ||
        (cpmac_priv->enable_vlan &&
        (length > 1518) &&
        !CPMAC_VLAN_IS_802_1Q_FRAME (skb->data))) {
      dev_kfree_skb_any(skb);
      if (length > 1518) {
        cpmac_priv->local_stats_rx_length_errors++;
      }
      cpmac_priv->local_stats_rx_errors++;
      DEB_WARN ("cpmac_if_data_from_phy, drop pkt len %u\n",
        length);
      ret = CPMAC_ERR_DROPPED;
    } else {
      /* Got VLAN packet. Relabel it. */
      if (CPMAC_VLAN_IS_802_1Q_FRAME (skb->data)) {
        unsigned short vid = CPMAC_VLAN_GET_VLAN_ID(skb->data);
		int devoffset = cpmac_priv->map_devoffset_in[vid & 0xf];
		int port = cpmac_priv->map_port_in[vid & 0xf];

		if ((p_dev = cpmac_priv->devs[devoffset]) == 0) {
		   DEB_ERR("cpmac: packet dropped, device disappeared\n");
		   dev_kfree_skb_any(skb);
		   return CPMAC_ERR_DROPPED;
		}

        skb->uniq_id &= ~(0xFF << 24);
        skb->uniq_id |= (port + 1) << 24;
        is_wan = (port == cpmac_priv->wanport);
#ifdef CPMAC_DMA_RX_PRIOQUEUE_DEBUG
        log_from_dmaqueue(is_wan);
#endif
		if (!is_wan || cpmac_priv->wanport_keeptags == 0) {
		   memmove(skb->data+VLAN_HLEN, skb->data, 12);
	       skb_pull(skb, VLAN_HLEN);
        }

#ifdef CONFIG_IP_MULTICAST_FASTFORWARD
	    if (cpmac_mcfw_netdrv) {
		   if (port == cpmac_priv->wanport) {
		      int rc;
			  rc = mcfw_multicast_forward_ethernet(cpmac_mcfw_sourceid, skb);
			  if (rc != 0) {
			     cpmac_priv->local_stats_rx_packets++;
			     cpmac_priv->local_stats_rx_bytes += length;
			     return CPMAC_ERR_NOERR;
			  }
		   } else {
              mcfw_snoop_recv(cpmac_mcfw_netdrv, port, skb);
		   }
	    }
#endif

      } else if (cpmac_priv->enable_vlan) {
        /*--- DEB_WARN("cpmac_if_data_from_phy, no 802.1, %u\n", skb->len); ---*/
        DEB_INFO("cpmac_if_data_from_phy, no 802.1, %u\n", skb->len);
      } else {
        /*
		 * HW 94.0.0.0
		 */
#ifdef CONFIG_IP_MULTICAST_FASTFORWARD
	    if (cpmac_mcfw_netdrv) {
		   if (mcfw_snoop_recv(cpmac_mcfw_netdrv, 0, skb) == 0) {
			  int rc;
		      rc = mcfw_multicast_forward_ethernet(cpmac_mcfw_sourceid, skb);
			  if (rc != 0) {
			     cpmac_priv->local_stats_rx_packets++;
			     cpmac_priv->local_stats_rx_bytes += length;
			     return CPMAC_ERR_NOERR;
			  }
		   }
	    }
#endif
	  }
      DEB_TRC ("cpmac_if_data_from_phy, %u %18B ... \n",
               skb->len, skb->data);
      skb->protocol = eth_type_trans (skb, p_dev);
	  skb->dev = p_dev;
	  if (is_wan && cpmac_wan_receive)
		 (*cpmac_wan_receive)(skb);
	  else
	     netif_rx (skb);
      cpmac_priv->local_stats_rx_packets++;
      cpmac_priv->local_stats_rx_bytes += length;
    }
  }
  return ret;
}
/*--- EXPORT_SYMBOL(cpmac_if_data_from_phy); ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_if_data_to_phy_complete (cpmac_priv_t *cpmac_priv, struct sk_buff *skb, 
    unsigned int status)
{
  cpmac_err_t ret = CPMAC_ERR_NOERR;

  if ((ret = cpmac_if_check_mac_handle (cpmac_priv)) != CPMAC_ERR_NOERR) {
    DEB_ERR("cpmac_if_data_to_phy_complete, ill mac_handle, %p\n",
        cpmac_priv);
  } else if (status == CPMAC_ERR_NOERR) {
    DEB_INFO("cpmac_if_data_to_phy_complete, %u bytes sent\n",
        skb->len);
    cpmac_priv->local_stats_tx_packets++;
    cpmac_priv->local_stats_tx_bytes += skb->len;
  } else {
    DEB_INFO("cpmac_if_data_to_phy_complete, %u bytes dropped (%u)\n",
            skb->len, status);
    cpmac_priv->local_stats_tx_errors++;
  }

#ifdef CPMAC_SIGNAL_CONGESTION
  if (status != CPMAC_ERR_DROPPED) {
    cpmac_priv->TxComplete++;
    if (!cpmac_priv->TxProbe) {
      if ((cpmac_priv->TxCount - cpmac_priv->TxComplete) == (CPPHY_MAX_TX_BUFFERS - 1)) {
        netif_wake_queue (cpmac_priv->owner);
      }
    } else {
      cpmac_priv->TxProbeComplete = 1;
    }
  }
#endif

  skb_trace(skb, 20);
  dev_kfree_skb_any (skb);
  return ret;
}
/*--- EXPORT_SYMBOL(cpmac_if_data_to_phy_complete); ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#ifdef CPMAC_TX_TIMEOUT
cpmac_err_t cpmac_if_teardown_complete (cpmac_priv_t *cpmac_priv)
{
  DEB_WARN ("cpmac_if_teardown_complete, init\n");
  netif_wake_queue (cpmac_priv->owner);
  return CPMAC_ERR_NOERR;
}
/*--- EXPORT_SYMBOL(cpmac_if_teardown_complete); ---*/
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_if_register (cpmac_phy_handle_t phy_handle, unsigned int phy_id, cpmac_service_funcs_t *service_funcs)
{
  struct net_device *p_dev;
  cpmac_priv_t *cpmac_priv;
  int failed;
  cpmac_err_t ret = CPMAC_ERR_NOERR;
  char DevName[16];

  if (phy_id >= CPMAC_MAX_PHY) {
    DEB_ERR("cpmac_if_register, phy_id %u exceeds limit\n",
        phy_id);
    ret = CPMAC_ERR_EXCEEDS_LIMIT;
  } else if (g_dev_array[phy_id].p_dev) {
    DEB_ERR("cpmac_if_register, phy_id %u already registered\n",
        phy_id);
    ret = CPMAC_ERR_ALREADY_REGISTERED;
  } else {
    if (!((cpphy_global_t *)phy_handle)->cpmac_switch) {
      sprintf (DevName, "eth%u", phy_id); /* Normal PHY */
    } else {
      sprintf (DevName, "cpmac%u", phy_id); /* Switch */
    }
    if (!(p_dev = alloc_netdev (sizeof (cpmac_priv_t),
      DevName, cpmac_main_dev_init))) {
      DEB_ERR("cpmac_if_register, no mem for device.\n");
      ret = CPMAC_ERR_NOMEM;
    } else {
      cpmac_priv = (cpmac_priv_t*)netdev_priv (p_dev);
      cpmac_priv->owner = p_dev;
      cpmac_priv->inst = phy_id;
      /*--- cpmac_priv->cpu_freq = avalanche_clkc_get_freq(CLKC_MIPS); ---*/
      g_dev_array[cpmac_priv->inst].p_dev = p_dev;
      memcpy (&g_dev_array[cpmac_priv->inst].service_funcs, service_funcs, sizeof (cpmac_service_funcs_t));
      g_dev_array[cpmac_priv->inst].phy_handle = phy_handle;

      /* initialize the timers for the net device */
      if (cpmac_main_timer_init () == -1) {
          DEB_ERR("cpmac_if_register, failed to set timer(s) for %s.\n", p_dev->name);
      }
      clear_bit (0, &cpmac_priv->set_to_close);

      cpmac_devices_installed++;

      /* init PHY */
      g_dev_array[cpmac_priv->inst].service_funcs.init (g_dev_array[cpmac_priv->inst].phy_handle, 
          (void *)cpmac_priv);
      cpmac_main_start_timer ();

      tasklet_init(&cpmac_priv->tasklet, g_dev_array[cpmac_priv->inst].service_funcs.isr_tasklet, 
          (unsigned long)phy_handle);

      failed = register_netdev(p_dev);
      if (failed) {
        DEB_ERR("Could not register device for inst %d because of reason code %d.\n", 
            phy_id, failed);
        free_netdev (p_dev);
        ret = CPMAC_ERR_REGISTER_FAILED;
      } else {
        DEB_ERR("cpmac_if_register, dev %s (phy_id %u) registered\n",
            DevName, phy_id);
        (void)cpmac_set_queue(p_dev);
      }
    }
  }
  return ret;
}
/*--- EXPORT_SYMBOL(cpmac_if_register); ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_if_release (cpmac_priv_t *cpmac_priv)
{
  cpmac_err_t ret = CPMAC_ERR_NOERR;

  if ((ret = cpmac_if_check_mac_handle (cpmac_priv)) != CPMAC_ERR_NOERR) {
    DEB_ERR("cpmac_if_release, ill mac_handle, %p\n",
        cpmac_priv);
  } else {
    cpphy_func_deinit_t deinit;
    cpmac_phy_handle_t *phy_handle;
    struct net_device *p_dev = cpmac_priv->owner;

    DEB_INFO("cpmac_if_release, %s\n", p_dev->name);

    cpmac_main_stop_timer ();

    phy_handle = g_dev_array[cpmac_priv->inst].phy_handle;
    deinit = g_dev_array[cpmac_priv->inst].service_funcs.deinit;

    netif_carrier_off(p_dev);
    unregister_netdev(p_dev);   /* this will invoke cpmac_main_dev_close(): teardown */

    /* leave int working until teardown has completed */
    set_bit (0, &cpmac_priv->set_to_close); 
    tasklet_kill (&cpmac_priv->tasklet);

    g_dev_array[cpmac_priv->inst].p_dev = NULL;
    deinit (phy_handle);

    kfree(p_dev);

    if (cpmac_devices_installed) {
      cpmac_devices_installed--;
    }
  }
  return ret;
}
EXPORT_SYMBOL(cpmac_if_release);
