/*------------------------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*------------------------------------------------------------------------------------------*/
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <net/pkt_sched.h>
#include <asm/mips-boards/ohioint.h>
#include <asm/mips-boards/prom.h>
#include <asm/mach-ohio/led_hal.h>
#include <asm/mach_avm.h>
#include <asm/io.h>
#include <asm/atomic.h>


#if !defined(CONFIG_NETCHIP_ADM69961)
#define CONFIG_NETCHIP_ADM69961
#endif
#include <linux/avm_event.h>
#include <linux/avm_cpmac.h>

#include "cpmac_if.h"
#include "cpmac_const.h"
#include "cpmac_types.h"
#include "cpmac_debug.h"
#include "cpmac_main.h"
#include "cpmac_reg.h"
#include "cpphy_const.h"
#include "cpphy_types.h"
#include "cpphy_mdio.h"
#include "cpphy_cppi.h"
#include "cpphy_main.h"
#include "cpphy_if.h"
#include "cpphy_misc.h"
#include "adm6996.h"
#include "cpphy_adm6996.h"

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_if_free_tcb(cpphy_cppi_t *cppi, cpphy_tcb_t *tcb) {
    /*--- FreeCount++; ---*/
    /*--- if((FreeCount & 0xfff) == 0) ---*/
        /*--- DEB_ERR("[avm cpmac test] FreeCount = %u\n", FreeCount); ---*/

    /*--- if(cppi->FreeCount++ > 4) { ---*/
        /*--- cppi->FreeCount = 0; ---*/
    /*--- local_irq_disable(); ---*/
        /*--- kfree(tcb); ---*/
    /*--- local_irq_enable(); ---*/
        /*--- return; ---*/
    /*--- } ---*/
    tcb->Next = NULL;
    cppi->TxLastFree->Next = tcb;
    cppi->TxLastFree = tcb;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpphy_tcb_t *cpphy_if_alloc_tcb(cpphy_cppi_t *cppi) {
    cpphy_tcb_t *tcb;

    /*--- AllocCount++; ---*/
    /*--- if((AllocCount & 0xfff) == 0) ---*/
        /*--- DEB_ERR("[avm cpmac test] AllocCount = %u\n", AllocCount); ---*/

    local_irq_disable();
    if(cppi->TxFirstFree->Next == NULL) {
        /* TODO Make freeing the allocated space possible */
        tcb = kmalloc(sizeof(cpphy_tcb_t), GFP_ATOMIC);
        if(((unsigned int) tcb) & 0xf)
            DEB_ERR("[avm cpmac test] Attention! Allocated tcb at unaligned address!\n");

#       if 0
        tcb = kmalloc(2 * sizeof(cpphy_tcb_t), GFP_ATOMIC);
        /* Align at 16 bytes */
        tcb = (cpphy_tcb_t *) (((unsigned int) tcb + 0xf) & ~0xf);
#       endif /*--- #if 0 ---*/
        local_irq_enable();
        return tcb;
    }
    /*--- atomic_sub(1, &cppi->FreeTcbCount); ---*/
    tcb = (cpphy_tcb_t *) cppi->TxFirstFree;
    cppi->TxFirstFree = cppi->TxFirstFree->Next;
    local_irq_enable();
    return tcb;
}

void cpphy_if_data_from_queues(cpphy_cppi_t *cppi);
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_if_tx_complete(cpphy_cppi_t *cppi, 
                          struct sk_buff *skb, 
                          unsigned int status) {
    cpmac_priv_t *cpmac_priv = cppi->cpmac_priv;
    /* FIXME Is it okay to comment this out? */
    /*--- if (!cpmac_priv || !g_dev_array[cpmac_priv->inst].p_dev) { ---*/
        /*--- return; ---*/
    /*--- } ---*/
    if(status == CPMAC_ERR_NOERR) {
        DEB_INFO("cpphy_if_tx_complete, %u bytes sent\n", skb->len);
        cpmac_priv->local_stats_tx_packets++;
        cpmac_priv->local_stats_tx_bytes += skb->len;
    } else {
        DEB_WARN("cpphy_if_tx_complete, %u bytes dropped (%u)\n", skb->len, status);
        cpmac_priv->local_stats_tx_errors++;
    }

#   ifdef CPMAC_SIGNAL_CONGESTION
    if(status != CPMAC_ERR_DROPPED) {
        cpmac_priv->TxComplete++;
        /* FIXME netif_wake_queue(cpmac_priv->owner); irgendwie beruecksichtigen */
    }
#   endif

    cppi->TxPrioQueues.q[skb->uniq_id >> 24].DMAFree++;

    skb_trace(skb, 20);
    dev_kfree_skb_any(skb);

	{
       struct net_device *p_dev = cpmac_priv->owner;
	   struct Qdisc *q = p_dev->qdisc;

	   if (netif_queue_stopped(p_dev)) {
	      netif_wake_queue(p_dev);
	   } else if (q->q.qlen) {
	      netif_schedule(p_dev);
	   }
	}

    cpphy_if_data_from_queues(cppi);
}


/*------------------------------------------------------------------------------------------*\
 * Garbage collect tx DMA queue                                                             *
\*------------------------------------------------------------------------------------------*/
void cpphy_if_gc_tx_queue(cpphy_cppi_t *cppi) {
    register cpphy_tcb_t *tcb;
    static atomic_t only_one = { counter: 0 };

    if(atomic_add_return(1, &only_one) != 1) {
        return;
    }
    while ((tcb = (cpphy_tcb_t *) cppi->TxFirst) != NULL) {
        dma_cache_inv((unsigned long)(void *)tcb, sizeof(cpphy_tcb_t));
        if (tcb->mode & CB_OWNERSHIP_BIT) {
            atomic_set(&only_one, 0);
            return;
        }
        if(tcb->skb != NULL)
            cpphy_if_tx_complete(cppi, (struct sk_buff *) tcb->skb, CPMAC_ERR_NOERR);
        tcb->skb = NULL;

        /*----------------------------------------------------------------------------------*\
         * The last entry stays in the queue. Therefor the LastTx pointer needs not to be
         * touched.
        \*----------------------------------------------------------------------------------*/
        if(tcb == cppi->TxLast) {
            atomic_set(&only_one, 0);
            return;
        }
        cppi->TxFirst = tcb->Next;
        /*--- cpphy_if_free_tcb(tcb); ---*/
    }
    atomic_set(&only_one, 0);
    return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
inline void cpphy_if_tcb_enqueue(cpphy_tcb_queue_t *queue, cpphy_tcb_t *tcb) {
    if(queue->Last != NULL) {
        queue->Last->Next = tcb;
    }
    queue->Last = tcb;
    if(queue->First == NULL) {
        queue->First = tcb;
    }
    queue->Free--;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
inline cpphy_tcb_t *cpphy_if_tcb_dequeue(cpphy_tcb_queue_t *queue) {
    cpphy_tcb_t *tcb;

    tcb = queue->First;
    if(queue->First != NULL) {
        queue->First = queue->First->Next;
        queue->Free++;
    }
    if(queue->First == NULL) {
        queue->Last = NULL;
    }
    return tcb;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_if_data_to_phy_dma (cpphy_cppi_t *cppi, cpphy_tcb_t *newtcb) {
    cpphy_tcb_t *TmpPrevEnqueue;
    cpmac_priv_t *cpmac_priv = cppi->cpmac_priv;
    unsigned int base = cpmac_priv->owner->base_addr;
    cpphy_tcb_t *tcb;

    tcb = cppi->TxPrevEnqueue->Next;

    /* Setup Tx mode and size */
    memcpy(tcb, newtcb, 16);
    tcb->skb = newtcb->skb;
    cpphy_if_free_tcb(cppi, newtcb);
    tcb->start_time = jiffies;

    /* prepare current buffer descriptor to be available for dma */
    dma_cache_wback_inv((unsigned long)(void *)tcb, 16);

    /* increment management ptr */
    TmpPrevEnqueue = cppi->TxPrevEnqueue;
    cppi->TxPrevEnqueue = tcb;

    /* dma may go on by chaining current tx descriptor */
    *((volatile unsigned int *) CPPHY_VIRT_TO_VIRT_NO_CACHE(&TmpPrevEnqueue->HNext)) =
        CPPHY_VIRT_TO_PHYS(tcb);

    /* If tx complete int handling has stopped, restart here. */
    /* It does not work to check CB_EOQ_BIT here and restart dma if bit is set: Bug of cpmac? */

    if (!cppi->TxDmaActive) {
      cppi->TxDmaActive = 1;
      /* write CPPI TX HDP */
      (CPMAC_TX_HDP (base, cppi->TxChannel)) = CPPHY_VIRT_TO_PHYS(tcb);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

#ifdef CPMAC_DMA_TX_PRIOQUEUE_DEBUG
static char buf[60];
static int bufidx = 0;

static void log_to_dmaqueue(int priority)
{
  if (bufidx == sizeof(buf)-1) {
	 buf[bufidx] = 0;
	 DEB_INFO("cpmac: tx %s\n", buf);
	 bufidx = 0;
  }
  buf[bufidx++] = '0'+ (priority % CPPHY_PRIO_QUEUES);
}

#endif

void cpphy_if_data_from_queues(cpphy_cppi_t *cppi) {
    cpphy_tcb_t *tcb;
    unsigned int priority, count;

    if(atomic_add_return(1, &cppi->dequeue_running) != 1) {
        DEB_WARN("[avm cpmac test] cpphy_if_data_from_queues conflict\n"); /* FIXME */
        return;
    }

    /* Part 2: Find next skb from appropriate queue and send it */
    for( ;; ) {
        count = 0;

        for(priority = 0; priority < CPPHY_PRIO_QUEUES; priority++) {
            if(cppi->TxPrioQueues.q[priority].DMAFree == 0) {
                /*--- DEB_ERR("[avm cpmac test] DMA queue full\n"); ---*/
                continue;
            }
            if(cppi->TxPrioQueues.q[priority].Free == cppi->TxPrioQueues.q[priority].MaxSize) {
                continue;
            }
            count++;
            if(!cppi->TxPrioQueues.q[priority].Pause--) {
                tcb = cpphy_if_tcb_dequeue(&cppi->TxPrioQueues.q[priority]);
                if(tcb != NULL) {
                    cppi->TxPrioQueues.q[priority].DMAFree--;
#ifdef CPMAC_DMA_TX_PRIOQUEUE_DEBUG
                    log_to_dmaqueue(priority);
#endif
                    cpphy_if_data_to_phy_dma(cppi, tcb);
                }
                cppi->TxPrioQueues.q[priority].Pause = cppi->TxPrioQueues.q[priority].PauseInit;
            }
        }
        if((count == 0) && (atomic_sub_return(1, &cppi->dequeue_running) == 0)) {
            break;
        }
    }
            
    /* Part 3 (optional): Reschedule queue priorities */
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpphy_if_data_to_phy(cpmac_phy_handle_t phy_handle, struct sk_buff *skb) {
    cpphy_cppi_t *cppi = &((cpphy_global_t *)phy_handle)->cppi;
    cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;
    unsigned int priority, frame_length;
    cpphy_tcb_t *tcb;

    if(!cppi->TxOpen || cppi->TxTeardownPending || (mdio->state != CPPHY_MDIO_ST_LINKED)) {
        DEB_INFO("cpphy_if_data_to_phy, chan closing or not opened (%u:%u)\n",
            cppi->hw_state, cppi->TxTeardownPending);
        return CPMAC_ERR_CHAN_NOT_OPEN;
    }

    /******************************************\
     * Part 1: Add skb(s) to correct queue(s) *
    \******************************************/

    /* Check which queue to use */
    priority = skb->uniq_id >> 24;

    /*--- if(cppi->TxPrioQueues.q[priority].Free != 0) { ---*/
    /* prepare tx data to be available for dma */
    dma_cache_wback_inv((unsigned long)skb->data, skb->len);    

    /* supply min ether frame size */
    frame_length = skb->len;
    if (frame_length < 60) {
        frame_length = 60;  /* + 4 byte hardware added fcs -> min frame length of 64 bytes */
    }

    if(cppi->TxPrioQueues.q[priority].Free == 0)
	  return CPMAC_ERR_NO_BUFFER;

    /* Allocate tcb, set it up with the skb data */
    tcb = cpphy_if_alloc_tcb(cppi);

    tcb->HNext    = 0;
    tcb->BufPtr   = CPPHY_VIRT_TO_PHYS((unsigned int *)skb->data);
    tcb->Off_BLen = frame_length;
    tcb->mode     = (frame_length | CB_SOF_BIT | CB_EOF_BIT | CB_OWNERSHIP_BIT);
    dma_cache_wback_inv((unsigned long)(void *)tcb, 16);
    tcb->Next     = NULL;
    tcb->skb      = skb;

    /* If queue got to be too long, dequeue and throw away oldest element */
    if(cppi->TxPrioQueues.q[priority].Free == 0) {
        cpphy_tcb_t *oldtcb;

        oldtcb = cpphy_if_tcb_dequeue(&cppi->TxPrioQueues.q[priority]);
        if(oldtcb == NULL)
            DEB_ERR("[avm cpmac test] Error: cpphy_if_data_to_phy oldtcb == NULL (prio = %u)\n", priority); /* FIXME */
        cpphy_if_tx_complete(cppi, (struct sk_buff *) oldtcb->skb, CPMAC_ERR_DROPPED);
        cpphy_if_free_tcb(cppi, oldtcb);
        /*--- DEB_ERR("[avm cpmac] Dropped packet from priority queue %u!\n", priority); ---*/
    }
    /* Enqueue tcb to the corresponding priority queue */
    cpphy_if_tcb_enqueue(&cppi->TxPrioQueues.q[priority], tcb);
    /*--- if(cppi->TxPrioQueues.q[priority].Free < cppi->TxPrioQueues.q[priority].MaxSize - 3) { ---*/
        /*--- DEB_ERR("[avm cpmac test] Queue %u contains more than three elements\n", priority); ---*/
    /*--- } ---*/

    /*--- } else { ---*/
        /*--- err = CPMAC_ERR_NO_BUFFER; ---*/
    /*--- } ---*/

    cpphy_if_data_from_queues(cppi);

    return CPMAC_ERR_NOERR;
}


/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
#if 1
#else /*--- #if 1 ---*/
cpmac_err_t cpphy_if_data_to_phy (cpmac_phy_handle_t phy_handle, struct sk_buff *skb)
{
    cpphy_tcb_t *tcb;
    cpmac_priv_t *cpmac_priv = ((cpphy_global_t *)phy_handle)->cpmac_priv;
    cpphy_cppi_t *cppi = &((cpphy_global_t *)phy_handle)->cppi;
    cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;
    register cpphy_tcb_t *last = (cpphy_tcb_t *) cppi->TxLast;
    unsigned int base = cpmac_priv->owner->base_addr;
    unsigned int frame_length, dma_has_stopped;

    if (!cppi->TxOpen || cppi->TxTeardownPending || (mdio->state != CPPHY_MDIO_ST_LINKED)) {
        DEB_WARN("[cpphy_if_data_to_phy] chan closing or not opened (%u:%u)\n", cppi->hw_state, cppi->TxTeardownPending);
        return CPMAC_ERR_CHAN_NOT_OPEN;
    }

    /* prepare tx data to be available for dma */
    dma_cache_wback_inv((unsigned long)skb->data, skb->len);    

    /* supply min ether frame size */
    frame_length = skb->len;
    if (frame_length < 60) {
        frame_length = 60;  /* + 4 byte hardware added fcs -> min frame length of 64 bytes */
    }

    /*--- tcb = cpphy_if_alloc_tcb(); ---*/
    tcb = (cpphy_tcb_t *) cppi->TxLast->Next;

    tcb->HNext    = 0;
    tcb->BufPtr   = CPPHY_VIRT_TO_PHYS((unsigned int *)skb->data);
    tcb->Off_BLen = frame_length;
    tcb->mode     = (frame_length | CB_SOP_BIT | CB_EOP_BIT | CB_OWNERSHIP_BIT);
    /*--- tcb->Next     = NULL; ---*/
    tcb->skb      = skb;

    if(last == NULL) {
        /*----------------------------------------------------------------------------------*\
         * Startup case   TODO possibly preset a packet in the init function
        \*----------------------------------------------------------------------------------*/
        DEB_ERR("[avm cpmac test] Error! Wrong place! Should not be called!\n"); /* FIXME */
        /*--- if(cppi->TxFirst) ---*/
            /*--- panic("[cpphy_if_data_to_phy]: First/Last Error\n"); ---*/
        /*--- TxLast = TxFirst = last = tcb; ---*/
        /*--- dma_has_stopped = 1; ---*/  /* auf jeden fall: Startup ! */
        /*--- dma_cache_wback_inv((unsigned long)(void *)TxLast, 16); ---*/
    } else {
        /*----------------------------------------------------------------------------------*\
         *  "normal case"
         *
         *  wenn Last schon abgearbeitet ist muss das EOQ bit gesetzt sein, 
         *  der DMA Controller arbeitet nicht mehr !    dma_has_stopped = 1
        \*----------------------------------------------------------------------------------*/
        dma_cache_inv((unsigned long)(void *)last, 16);
        dma_has_stopped = last->mode & CB_OWNERSHIP_BIT ? 0 : 1;

        if(dma_has_stopped && !(last->mode & CB_EOQ_BIT)) {
            DEB_ERR("[cpphy_if_data_to_phy] last is end but no EOQ bit set\n");
        }
        /*----------------------------------------------------------------------------------*\
         * an den bestehenden last unser paket anhaengen
        \*----------------------------------------------------------------------------------*/
        last->HNext = CPPHY_VIRT_TO_PHYS(tcb);
        dma_cache_wback_inv((unsigned long)(void *)last, 16);  /* zurueckschreiben */
        /*--- last->Next  = tcb; ---*/
        /*--- hierdurch wird u.U. die Werte im Mode zerstoert ---*/
        cppi->TxLast = tcb;

        if(dma_has_stopped == 0) {  /* wir sollten eigentlich laufen */
            /*--- waren wir bein anhaengen zu langsam ? ---*/
            if((CPMAC_TX_HDP (base, cppi->TxChannel)) != 0) {
                /*--- TxLast = tcb; ---*/
                cpphy_if_gc_tx_queue(cppi);
                return CPMAC_ERR_NOERR;
            }
            last->mode  = CB_SOP_BIT | CB_EOP_BIT;
            dma_cache_wback_inv((unsigned long)(void *)last, 16);  /* zurueckschreiben */
        }
    }
    cpphy_if_gc_tx_queue(cppi);
    (CPMAC_TX_HDP (base, cppi->TxChannel)) = CPPHY_VIRT_TO_PHYS(cppi->TxFirst);

    return CPMAC_ERR_NOERR;
}
#endif /*--- #else ---*/ /*--- #if 1 ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if 0
int cpphy_cppi_tx_int (cpphy_cppi_t *cppi)
{
  cpphy_tcb_t *tcb, *tcb_ack, *tcb_last;
  unsigned int mode;
  unsigned int base = cppi->cpmac_priv->owner->base_addr;

  /* Handle teardown interrupt */
  if (cppi->TxTeardownPending) {
    if (cpphy_cppi_tx_teardown_int (cppi) == CPMAC_ERR_NOERR) {
      return CPMAC_ERR_NOERR;
    }
  }
  tcb = cppi->TxCurrDequeue;
  tcb_ack = (cpphy_tcb_t *)CPPHY_PHYS_TO_VIRT_CACHE(CPMAC_TX_INT_ACK (base, cppi->TxChannel));
  dma_cache_inv((unsigned long)(void *)tcb_ack, 16);
  /*--- mode = *((volatile unsigned int *) CPPHY_VIRT_TO_VIRT_NO_CACHE(&tcb_ack->mode)); ---*/
  if (!(tcb_ack->mode & CB_OWNERSHIP_BIT)) {
    /* Write the completion pointer */
    (CPMAC_TX_INT_ACK (base, cppi->TxChannel)) = CPPHY_VIRT_TO_PHYS(tcb_ack);

    /* if there is an error trace concerning locking: start here */
    if (*((volatile unsigned int *)&tcb_ack->HNext)) {
      /* check possible race condition with dma */
      if (tcb_ack->mode & CB_EOQ_BIT) {
        /* restart dma if halted */
        (CPMAC_TX_HDP (base, cppi->TxChannel)) = tcb_ack->HNext;
      }
    } else {
      cppi->TxDmaActive = 0;    /* signal cpphy_if_data_to_phy() to restart dma */
      if (*((volatile unsigned int *)&tcb_ack->HNext)) {
        /* this hopefully never happens, because of higher prio of tasklet */
        DEB_ERR("cpphy_cppi_tx_int, need locking with cpphy_if_data_to_phy()\n");
      }
    }
    /* if there is an error trace concerning locking: end here */

    do {
      cpmac_if_data_to_phy_complete (cppi->cpmac_priv, (struct sk_buff *)tcb->skb, CPMAC_ERR_NOERR);
      tcb->skb = NULL; /* tcb may be reused immediately */
      tcb_last = tcb;
      tcb = tcb->Next;
    } while (tcb_last != tcb_ack);
    cppi->TxCurrDequeue = tcb;
  } else {
    /* should never happen: tx int while no buffer completed */
    DEB_ERR("cpphy_cppi_tx_int, ill tcb_ack: %p (%p,%p)\n",
        tcb_ack, cppi->TxFirst, cppi->TxLast);
  }
}
#endif

/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
cpmac_err_t cpphy_if_control_req (cpmac_phy_handle_t phy_handle,
    cpmac_control_req_t control, ...)
{
  va_list param;
  cpmac_err_t ret = CPMAC_ERR_NOERR;

  switch (control) {
    case CPMAC_CONTROL_REQ_UPDATE:
      {
        struct cpmac_event_struct *event_struct;
        cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;

        mdio->event_notify = 0;
        mdio->event_last_update = 0;

        va_start (param, control);
        event_struct = va_arg (param, void *); 
        cpphy_mdio_update (mdio, event_struct);
        va_end (param);
      }
      break;
    case CPMAC_CONTROL_REQ_UPDATE_DETAIL:
      {
        struct cpmac_event_struct_verbose *event_struct;
        cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;

        va_start (param, control);
        event_struct = va_arg (param, void *); 
        cpphy_mdio_update_detail (mdio, event_struct);
        va_end (param);
      }
      break;
    case CPMAC_CONTROL_REQ_IS_SWITCH:
      if (!((cpphy_global_t *)phy_handle)->cpmac_switch) {
        ret = CPMAC_ERR_NO_SWITCH;
      }
      break;
    case CPMAC_CONTROL_REQ_MULTI_SINGLE:
      cpphy_cppi_set_multi_promiscous (&((cpphy_global_t *)phy_handle)->cppi,
          2, 0);
      break;
    case CPMAC_CONTROL_REQ_MULTI_ALL:
      cpphy_cppi_set_multi_promiscous (&((cpphy_global_t *)phy_handle)->cppi,
          1, 0);
      break;
    case CPMAC_CONTROL_REQ_PROMISCOUS:
      cpphy_cppi_set_multi_promiscous (&((cpphy_global_t *)phy_handle)->cppi,
          0, 1);
      break;
    case CPMAC_CONTROL_REQ_HW_STATUS:
      {
        struct net_device_stats *stats;
        cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;
        cpphy_cppi_t *cppi = &((cpphy_global_t *)phy_handle)->cppi;

        va_start (param, control);
        stats = va_arg (param, void *); 
        memset (stats, 0, sizeof (*stats));
        cpphy_mdio_update_hw_status (mdio, stats);
        cpphy_cppi_update_hw_status (cppi, stats);
        va_end (param);
      }
      break;
    case CPMAC_CONTROL_REQ_TEARDOWN:
      {
        unsigned int mode;
        cpphy_cppi_t *cppi = &((cpphy_global_t *)phy_handle)->cppi;

        va_start (param, control);
        mode = va_arg (param, unsigned int);
        ret = cpphy_cppi_teardown (cppi, mode);
        va_end (param);
      }
      break;
    case CPMAC_CONTROL_REQ_START_DMA:
      ret = cpphy_cppi_start_dma (&((cpphy_global_t *)phy_handle)->cppi);
      break;
    case CPMAC_CONTROL_REQ_PORT_COUNT:
      /* to check: adjust port count in case of empty cpphy */
      /* design: abuse ret as port count */
      ret = (cpmac_err_t)((cpphy_global_t *)phy_handle)->cpmac_switch ? 4 : 1;
      break;
    case CPMAC_CONTROL_REQ_SET_CONFIG:
      {
        cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;
        struct avm_cpmac_config_struct *info;

        va_start (param, control);
        info = va_arg (param, void *); 
        if (((cpphy_global_t *)phy_handle)->cpmac_switch) {
          ret = adm_configure_cpmac (mdio, info);
        } else {
          ret = CPMAC_ERR_NO_SWITCH;
        }
        va_end (param);
      }
      break;
    case CPMAC_CONTROL_REQ_GET_CONFIG:
      {
        cpphy_mdio_t *mdio = &((cpphy_global_t *)phy_handle)->mdio;
        struct avm_cpmac_config_struct *info;

        va_start (param, control);
        info = va_arg (param, void *); 
        if (((cpphy_global_t *)phy_handle)->cpmac_switch) {
          ret = adm_get_config_cpmac (mdio, info);
        } else {
          ret = CPMAC_ERR_NO_SWITCH;
        }
        va_end (param);
      }
      break;
    default:
      DEB_INFO("cpphy_control_req, unhandled control %u\n",
          control);
      ret = CPMAC_ERR_ILL_CONTROL;
      break;
  }
  return ret;
}

/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
void cpphy_if_isr_end(cpmac_phy_handle_t phy_handle)
{
  CPMAC_MAC_EOI_VECTOR (((cpphy_global_t *)phy_handle)->cpmac_priv->owner->base_addr) = 0;
}

/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
void cpphy_if_isr_tasklet (unsigned long context)
{
  cpmac_phy_handle_t phy_handle = (cpmac_phy_handle_t)context;
  unsigned int IntVec;
  cpphy_cppi_t *cppi = &((cpphy_global_t *)phy_handle)->cppi;
  cpmac_priv_t *cpmac_priv = ((cpphy_global_t *)phy_handle)->cpmac_priv;

  if (cpmac_priv->DoTick) {
    cpmac_priv->DoTick = 0;
    cpphy_mdio_tick (&((cpphy_global_t *)phy_handle)->mdio);
  }

  /* Verify proper device state - important because a call prior to Open would
      result in a lockup */
  if (cppi->hw_state != CPPHY_HW_ST_OPENED)
     return;

  IntVec = CPMAC_MAC_IN_VECTOR (cpmac_priv->owner->base_addr);

  DEB_TRC("cpphy_if_isr_tasklet, int vec 0x%X\n",
    IntVec);

/*--- #if 0 ---*/
  if (IntVec & MAC_IN_VECTOR_TX_INT_OR) { /* only chan 0 in use */
    /*--- if (IntVec & 0x7) { ---*/
      /*--- DEB_ERR("cpphy_if_isr_tasklet, unexpected tx chan %u\n", ---*/
        /*--- IntVec & 0x7); ---*/
    /*--- } else { ---*/
      cpphy_cppi_tx_int (cppi);
    /*--- } ---*/
  }
/*--- #endif ---*/ 

  if (IntVec & MAC_IN_VECTOR_RX_INT_OR) { /* only chan 0 in use */
    if ((IntVec >> 8) & 0x7) {
      DEB_ERR("cpphy_if_isr_tasklet, unexpected rx chan %u\n",
          (IntVec >> 8) & 0x7);
    } else {
      cpphy_cppi_rx_int (cppi);
    }
  }

  if (IntVec & MAC_IN_VECTOR_HOST_INT) {
    /* severe problem */
    DEB_ERR("cpphy_if_isr_tasklet, status %08X for base:%X\n",
      CPMAC_MACSTATUS (cpmac_priv->owner->base_addr),
      (unsigned int)cpmac_priv->owner->base_addr);
    /* to do: reset hardware */
  }
  if (IntVec & (MAC_IN_VECTOR_TX_INT_OR | MAC_IN_VECTOR_RX_INT_OR | MAC_IN_VECTOR_HOST_INT)) {
    CPMAC_MAC_EOI_VECTOR (cpmac_priv->owner->base_addr) = 0;
  }
}

/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
cpmac_err_t cpphy_if_init (cpmac_phy_handle_t phy_handle, cpmac_priv_t *cpmac_priv)
{
  cpmac_err_t ret = CPMAC_ERR_NOERR;
  struct net_device *p_dev = cpmac_priv->owner;
  char *mac_name          = NULL;
  char *mac_string        = NULL;

  ((cpphy_global_t *)phy_handle)->cpmac_priv = cpmac_priv;
  ((cpphy_global_t *)phy_handle)->cppi.cpmac_priv = cpmac_priv;

  /* reset of cpmac and phy */
  if (((cpphy_global_t *)phy_handle)->high_phy) {
    cpmac_priv->mac_reset_bit = EMACB_RESET_BIT;
    /* to check: reset allthough no internal phy on high cpmac */
    cpmac_priv->phy_reset_bit = -1; /* signal: no bit to reset */
    cpmac_priv->mac_pdown_bit = PDCR_BIT_EMAC1;
    cpmac_priv->intr = OHIOINT_CPMAC1;
    p_dev->if_port = CPPHY_CPMAC_HIGH_PORT_ID;            
    p_dev->base_addr = OHIO_CPMAC1_BASE;
    mac_name="macb";
  } else {
    cpmac_priv->mac_reset_bit = EMACA_RESET_BIT;
    cpmac_priv->phy_reset_bit = EMAC_PHY_RESET_BIT;
    cpmac_priv->mac_pdown_bit = PDCR_BIT_EMAC0;
    cpmac_priv->intr = OHIOINT_CPMAC0;
    p_dev->if_port = CPPHY_CPMAC_LOW_PORT_ID;            
    p_dev->base_addr = OHIO_CPMAC0_BASE;
    mac_name="maca";
  }

  avm_reset_device (cpmac_priv->mac_reset_bit, 100);
  if (cpmac_priv->phy_reset_bit != -1) {
      avm_reset_device (cpmac_priv->phy_reset_bit, 100);
  }
  /*--- wait 100 ms ---*/
  set_current_state(TASK_INTERRUPTIBLE);
  schedule_timeout(HZ / 10);
  avm_take_device_out_of_power_down(cpmac_priv->mac_pdown_bit);

  mac_string = prom_getenv(mac_name);
  if (!mac_string) {
      mac_string="08.00.28.32.06.02";
      DEB_ERR("Error getting mac from Boot enviroment for %s\n", p_dev->name);
      DEB_ERR("Using default mac address: %s\n",mac_string);
      DEB_ERR("Use Bootloader command:\n");
      DEB_ERR("    setenv %s xx.xx.xx.xx.xx.xx\n",mac_name);
      DEB_ERR("to set mac address\n");
  }
  DEB_INFO("cpphy_if_init, dev %s has mac addr: %s\n",
      p_dev->name, mac_string);
  cpphy_misc_str2eaddr (p_dev->dev_addr, mac_string);

  /* Use the external PHY only for the switch variants if the FBox */
  if (((cpphy_global_t *)phy_handle)->cpmac_switch &&
      ((cpphy_global_t *)phy_handle)->is_ohio) {
    /* Reset EPHY, leave it in reset state, because we use the external PHY */
    if (cpmac_priv->phy_reset_bit != -1) {
      avm_put_device_into_reset (cpmac_priv->phy_reset_bit);
    }

    /* Delay 200ms */
    set_current_state(TASK_INTERRUPTIBLE);
    schedule_timeout(HZ / 5);

    /* MII pins are connected to the MII interface on the EMAC0 module */
    *((volatile unsigned int *)OHIO_MII_SEL_REG) = 1;
  }

  /* Tasklet is initialized at the isr registeration time. */
  cpmac_priv->dev_size = 0x800;     /* to check */

  request_mem_region(p_dev->base_addr, cpmac_priv->dev_size, p_dev->name);

  cpphy_main_open ((cpphy_global_t *)phy_handle, p_dev);
  return ret;
}

/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
void cpphy_if_deinit (cpmac_phy_handle_t phy_handle)
{
  cpmac_priv_t *cpmac_priv = ((cpphy_global_t *)phy_handle)->cpmac_priv;
  struct net_device *p_dev = (struct net_device *)(cpmac_priv->owner);
  cpphy_cppi_t *cppi = &((cpphy_global_t *)phy_handle)->cppi;

  DEB_TRC("cpphy_if_deinit, init\n");

  if (cpphy_main_close(cppi) != CPMAC_ERR_NOERR) {
      DEB_ERR("cpphy_if_deinit, failed to teardown channel\n");
  } else {
      /* hope that the HAL closes down the tick timer.*/
      DEB_INFO("cpphy_if_deinit, device %s closed\n", p_dev->name);
  }

  /* Buffer/descriptor resources may still need to be freed if a Close
    Mode 1 was performed prior to Shutdown - clean up here */
  if (cppi->RcbStart) {
    cpphy_cppi_free_rcb (cppi);
  }
  if (cppi->TcbStart) {
    cpphy_cppi_free_tcb (cppi);
  }

  release_mem_region(p_dev->base_addr, cpmac_priv->dev_size);
}
